/* Generated by Cython 0.29.21 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__mpfmc__uix__bitmap_font__bitmap_font
#define __PYX_HAVE_API__mpfmc__uix__bitmap_font__bitmap_font
/* Early includes */
#include <string.h>
#include "SDL.h"
#include "SDL_image.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 1
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING "utf-8"
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "stringsource",
  "mpfmc\\uix\\bitmap_font\\bitmap_font.pyx",
};

/*--- Type declarations ---*/
struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter;
struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont;
struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer;

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":614
 * 
 * 
 * cdef class BitmapFontCharacter:             # <<<<<<<<<<<<<<
 *     cdef public int id
 *     cdef public SDL_Rect rect
 */
struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter {
  PyObject_HEAD
  int id;
  struct SDL_Rect rect;
  int xoffset;
  int yoffset;
  int xadvance;
};


/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":632
 * 
 * 
 * cdef class BitmapFont:             # <<<<<<<<<<<<<<
 *     cdef public str face
 *     cdef public bint bold
 */
struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont {
  PyObject_HEAD
  PyObject *face;
  int bold;
  int italic;
  int padding;
  int spacing;
  int outline;
  int line_height;
  int base;
  int scale_w;
  int scale_h;
  PyObject *characters;
  PyObject *kernings;
  struct SDL_Surface *image;
};


/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":928
 * 
 * 
 * cdef class _SurfaceContainer:             # <<<<<<<<<<<<<<
 *     cdef SDL_Surface* surface
 *     cdef int w, h
 */
struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer {
  PyObject_HEAD
  struct SDL_Surface *surface;
  int w;
  int h;
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* UnicodeAsUCS4.proto */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject*);

/* object_ord.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Ord(c)\
    (likely(PyUnicode_Check(c)) ? (long)__Pyx_PyUnicode_AsPy_UCS4(c) : __Pyx__PyObject_Ord(c))
#else
#define __Pyx_PyObject_Ord(c) __Pyx__PyObject_Ord(c)
#endif
static long __Pyx__PyObject_Ord(PyObject* c);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* unicode_iter.proto */
static CYTHON_INLINE int __Pyx_init_unicode_iteration(
    PyObject* ustring, Py_ssize_t *length, void** data, int *kind);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

static PyObject* __pyx_convert__to_py_struct__SDL_Rect(struct SDL_Rect s);
/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'mpfmc.uix.bitmap_font.bitmap_font' */
static PyTypeObject *__pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter = 0;
static PyTypeObject *__pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont = 0;
static PyTypeObject *__pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer = 0;
static struct SDL_Rect __pyx_convert__from_py_struct__SDL_Rect(PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "mpfmc.uix.bitmap_font.bitmap_font"
extern int __pyx_module_is_main_mpfmc__uix__bitmap_font__bitmap_font;
int __pyx_module_is_main_mpfmc__uix__bitmap_font__bitmap_font = 0;

/* Implementation of 'mpfmc.uix.bitmap_font.bitmap_font' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_h[] = "h";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k__3[] = "";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_char[] = "char";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_find[] = "find";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_rect[] = "rect";
static const char __pyx_k_rgba[] = "rgba";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_text[] = "text";
static const char __pyx_k_chars[] = "chars";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_first[] = "first";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_group[] = "group";
static const char __pyx_k_width[] = "width";
static const char __pyx_k_amount[] = "amount";
static const char __pyx_k_attrib[] = "attrib";
static const char __pyx_k_common[] = "common";
static const char __pyx_k_height[] = "height";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_isfile[] = "isfile";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_search[] = "search";
static const char __pyx_k_second[] = "second";
static const char __pyx_k_findall[] = "findall";
static const char __pyx_k_getroot[] = "getroot";
static const char __pyx_k_kerning[] = "kerning";
static const char __pyx_k_options[] = "options";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_xoffset[] = "xoffset";
static const char __pyx_k_yoffset[] = "yoffset";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_kernings[] = "kernings";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_xadvance[] = "xadvance";
static const char __pyx_k_ImageData[] = "ImageData";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_container[] = "container";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_get_image[] = "get_image";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_BitmapFont[] = "BitmapFont";
static const char __pyx_k_IGNORECASE[] = "IGNORECASE";
static const char __pyx_k_ParseError[] = "ParseError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_descriptor[] = "descriptor";
static const char __pyx_k_image_file[] = "image_file";
static const char __pyx_k_lineHeight[] = "lineHeight";
static const char __pyx_k_startswith[] = "startswith";
static const char __pyx_k_ElementTree[] = "ElementTree";
static const char __pyx_k_bitmap_font[] = "bitmap_font";
static const char __pyx_k_base_0_9_1_5[] = "base=([0-9]{1,5})";
static const char __pyx_k_font_kerning[] = "font_kerning";
static const char __pyx_k_get_kernings[] = "get_kernings";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_get_characters[] = "get_characters";
static const char __pyx_k_get_font_asset[] = "get_font_asset";
static const char __pyx_k_kivy_core_image[] = "kivy.core.image";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_SurfaceContainer[] = "_SurfaceContainer";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_lineHeight_0_9_1_5[] = "lineHeight=([0-9]{1,5})";
static const char __pyx_k_BitmapFontCharacter[] = "BitmapFontCharacter";
static const char __pyx_k_BitmapFontException[] = "BitmapFontException";
static const char __pyx_k_load_descriptor_xml[] = "_load_descriptor_xml";
static const char __pyx_k_load_descriptor_file[] = "_load_descriptor_file";
static const char __pyx_k_load_descriptor_list[] = "_load_descriptor_list";
static const char __pyx_k_xml_etree_ElementTree[] = "xml.etree.ElementTree";
static const char __pyx_k_kerning_s_first_P_first_0_9_1_3[] = "kerning\\s+first=(?P<first>[0-9]{1,3})\\s+second=(?P<second>[0-9]{1,3})\\s+amount=(?P<amount>-?[0-9]{1,3})";
static const char __pyx_k_Bitmap_font_descriptor_file_inva[] = "Bitmap font descriptor file invalid format";
static const char __pyx_k_Could_not_load_bitmap_font_image[] = "Could not load bitmap font image file";
static const char __pyx_k_Could_not_locate_the_bitmap_font[] = "Could not locate the bitmap font descriptor file ";
static const char __pyx_k_Exception_returned_by_the_bitmap[] = "Exception returned by the bitmap font module";
static const char __pyx_k_Illegal_value_in_bitmap_font_des[] = "Illegal value in bitmap font descriptor";
static const char __pyx_k_No_value_specified_for_struct_at[] = "No value specified for struct attribute 'x'";
static const char __pyx_k_char_s_id_P_id_0_9_1_4_s_x_P_x_0[] = "char\\s+id=(?P<id>[0-9]{1,4})\\s+x=(?P<x>[0-9]{1,4})\\s+y=(?P<y>[0-9]{1,4})\\s+width=(?P<width>[0-9]{1,3})\\s+height=(?P<height>[0-9]{1,3})\\s+xoffset=(?P<xoffset>-?[0-9]{1,3})\\s+yoffset=(?P<yoffset>-?[0-9]{1,3})\\s+xadvance=(?P<xadvance>[0-9]{1,3})\\s+page=(?P<page>[0-9]{1,2})\\s+chnl=(?P<chnl>[0-9]{1,3})";
static const char __pyx_k_mpfmc_uix_bitmap_font_bitmap_fon[] = "mpfmc.uix.bitmap_font.bitmap_font";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_Bitmap_font_descriptor_file_inva_2[] = "Bitmap font descriptor file invalid XML format";
static const char __pyx_k_No_value_specified_for_struct_at_2[] = "No value specified for struct attribute 'y'";
static const char __pyx_k_No_value_specified_for_struct_at_3[] = "No value specified for struct attribute 'w'";
static const char __pyx_k_No_value_specified_for_struct_at_4[] = "No value specified for struct attribute 'h'";
static PyObject *__pyx_n_s_BitmapFont;
static PyObject *__pyx_n_u_BitmapFont;
static PyObject *__pyx_n_s_BitmapFontCharacter;
static PyObject *__pyx_n_s_BitmapFontException;
static PyObject *__pyx_n_u_BitmapFontException;
static PyObject *__pyx_kp_u_Bitmap_font_descriptor_file_inva;
static PyObject *__pyx_kp_u_Bitmap_font_descriptor_file_inva_2;
static PyObject *__pyx_kp_u_Could_not_load_bitmap_font_image;
static PyObject *__pyx_kp_u_Could_not_locate_the_bitmap_font;
static PyObject *__pyx_n_s_ElementTree;
static PyObject *__pyx_kp_s_Exception_returned_by_the_bitmap;
static PyObject *__pyx_n_s_IGNORECASE;
static PyObject *__pyx_kp_u_Illegal_value_in_bitmap_font_des;
static PyObject *__pyx_n_s_ImageData;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at_2;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at_3;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at_4;
static PyObject *__pyx_n_s_ParseError;
static PyObject *__pyx_n_s_SurfaceContainer;
static PyObject *__pyx_n_u_SurfaceContainer;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_u_amount;
static PyObject *__pyx_n_s_attrib;
static PyObject *__pyx_n_u_base;
static PyObject *__pyx_kp_u_base_0_9_1_5;
static PyObject *__pyx_n_s_bitmap_font;
static PyObject *__pyx_n_u_char;
static PyObject *__pyx_kp_u_char_s_id_P_id_0_9_1_4_s_x_P_x_0;
static PyObject *__pyx_n_u_chars;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_u_common;
static PyObject *__pyx_n_s_container;
static PyObject *__pyx_n_s_descriptor;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_find;
static PyObject *__pyx_n_s_findall;
static PyObject *__pyx_n_u_first;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_font_kerning;
static PyObject *__pyx_n_u_font_kerning;
static PyObject *__pyx_n_s_get_characters;
static PyObject *__pyx_n_s_get_font_asset;
static PyObject *__pyx_n_s_get_image;
static PyObject *__pyx_n_s_get_kernings;
static PyObject *__pyx_n_s_getroot;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_group;
static PyObject *__pyx_n_s_h;
static PyObject *__pyx_n_u_height;
static PyObject *__pyx_n_u_id;
static PyObject *__pyx_n_s_image_file;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_u_info;
static PyObject *__pyx_n_s_isfile;
static PyObject *__pyx_n_u_kerning;
static PyObject *__pyx_kp_u_kerning_s_first_P_first_0_9_1_3;
static PyObject *__pyx_n_u_kernings;
static PyObject *__pyx_n_s_kivy_core_image;
static PyObject *__pyx_n_u_lineHeight;
static PyObject *__pyx_kp_u_lineHeight_0_9_1_5;
static PyObject *__pyx_n_s_load_descriptor_file;
static PyObject *__pyx_n_s_load_descriptor_list;
static PyObject *__pyx_n_s_load_descriptor_xml;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_mpfmc_uix_bitmap_font_bitmap_fon;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_options;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_re;
static PyObject *__pyx_n_s_rect;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_u_rgba;
static PyObject *__pyx_n_s_search;
static PyObject *__pyx_n_u_second;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_startswith;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_text;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_n_u_width;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_u_x;
static PyObject *__pyx_n_s_xadvance;
static PyObject *__pyx_n_u_xadvance;
static PyObject *__pyx_n_s_xml_etree_ElementTree;
static PyObject *__pyx_n_s_xoffset;
static PyObject *__pyx_n_u_xoffset;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_u_y;
static PyObject *__pyx_n_s_yoffset;
static PyObject *__pyx_n_u_yoffset;
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter___cinit__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kw); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont___cinit__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kw); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_2__init__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_image_file, PyObject *__pyx_v_descriptor); /* proto */
static void __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4__dealloc__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6get_image(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_8get_characters(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_10get_kernings(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_12_load_descriptor_list(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_descriptor_list); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_14_load_descriptor_file(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_descriptor_file); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_16_load_descriptor_xml(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_xml_tree); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_18get_descent(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_20get_ascent(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_22get_extents(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_font_kerning); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_4__del__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_24__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_26__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer___cinit__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self, PyObject *__pyx_v_w, PyObject *__pyx_v_h); /* proto */
static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_2__init__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self, PyObject *__pyx_v_w, PyObject *__pyx_v_h); /* proto */
static void __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_4__dealloc__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_6render(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self, PyObject *__pyx_v_container, PyObject *__pyx_v_text, PyObject *__pyx_v_x, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_8get_data(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_4;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
/* Late includes */

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":621
 *     cdef public int xadvance
 * 
 *     def __cinit__(self, *args, **kw):             # <<<<<<<<<<<<<<
 *         self.id = 0
 *         self.rect.x = 0
 */

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kw = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 1))) return -1;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter___cinit__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self), __pyx_v_args, __pyx_v_kw);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kw);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter___cinit__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kw) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":622
 * 
 *     def __cinit__(self, *args, **kw):
 *         self.id = 0             # <<<<<<<<<<<<<<
 *         self.rect.x = 0
 *         self.rect.y = 0
 */
  __pyx_v_self->id = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":623
 *     def __cinit__(self, *args, **kw):
 *         self.id = 0
 *         self.rect.x = 0             # <<<<<<<<<<<<<<
 *         self.rect.y = 0
 *         self.rect.w = 0
 */
  __pyx_v_self->rect.x = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":624
 *         self.id = 0
 *         self.rect.x = 0
 *         self.rect.y = 0             # <<<<<<<<<<<<<<
 *         self.rect.w = 0
 *         self.rect.h = 0
 */
  __pyx_v_self->rect.y = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":625
 *         self.rect.x = 0
 *         self.rect.y = 0
 *         self.rect.w = 0             # <<<<<<<<<<<<<<
 *         self.rect.h = 0
 *         self.xoffset = 0
 */
  __pyx_v_self->rect.w = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":626
 *         self.rect.y = 0
 *         self.rect.w = 0
 *         self.rect.h = 0             # <<<<<<<<<<<<<<
 *         self.xoffset = 0
 *         self.yoffset = 0
 */
  __pyx_v_self->rect.h = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":627
 *         self.rect.w = 0
 *         self.rect.h = 0
 *         self.xoffset = 0             # <<<<<<<<<<<<<<
 *         self.yoffset = 0
 *         self.xadvance = 0
 */
  __pyx_v_self->xoffset = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":628
 *         self.rect.h = 0
 *         self.xoffset = 0
 *         self.yoffset = 0             # <<<<<<<<<<<<<<
 *         self.xadvance = 0
 * 
 */
  __pyx_v_self->yoffset = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":629
 *         self.xoffset = 0
 *         self.yoffset = 0
 *         self.xadvance = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->xadvance = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":621
 *     cdef public int xadvance
 * 
 *     def __cinit__(self, *args, **kw):             # <<<<<<<<<<<<<<
 *         self.id = 0
 *         self.rect.x = 0
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":615
 * 
 * cdef class BitmapFontCharacter:
 *     cdef public int id             # <<<<<<<<<<<<<<
 *     cdef public SDL_Rect rect
 *     cdef public int xoffset
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->id); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 615, __pyx_L1_error)
  __pyx_v_self->id = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":616
 * cdef class BitmapFontCharacter:
 *     cdef public int id
 *     cdef public SDL_Rect rect             # <<<<<<<<<<<<<<
 *     cdef public int xoffset
 *     cdef public int yoffset
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert__to_py_struct__SDL_Rect(__pyx_v_self->rect); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.rect.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  struct SDL_Rect __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_convert__from_py_struct__SDL_Rect(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 616, __pyx_L1_error)
  __pyx_v_self->rect = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.rect.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":617
 *     cdef public int id
 *     cdef public SDL_Rect rect
 *     cdef public int xoffset             # <<<<<<<<<<<<<<
 *     cdef public int yoffset
 *     cdef public int xadvance
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->xoffset); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.xoffset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 617, __pyx_L1_error)
  __pyx_v_self->xoffset = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.xoffset.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":618
 *     cdef public SDL_Rect rect
 *     cdef public int xoffset
 *     cdef public int yoffset             # <<<<<<<<<<<<<<
 *     cdef public int xadvance
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->yoffset); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.yoffset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 618, __pyx_L1_error)
  __pyx_v_self->yoffset = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.yoffset.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":619
 *     cdef public int xoffset
 *     cdef public int yoffset
 *     cdef public int xadvance             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, *args, **kw):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->xadvance); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.xadvance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 619, __pyx_L1_error)
  __pyx_v_self->xadvance = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.xadvance.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2__reduce_cython__[] = "BitmapFontCharacter.__reduce_cython__(self)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2__reduce_cython__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4__setstate_cython__[] = "BitmapFontCharacter.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4__setstate_cython__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":650
 *     cdef SDL_Surface *image
 * 
 *     def __cinit__(self, *args, **kw):             # <<<<<<<<<<<<<<
 *         self.face = ""
 *         self.bold = False
 */

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kw = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 1))) return -1;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont___cinit__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), __pyx_v_args, __pyx_v_kw);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kw);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont___cinit__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kw) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":651
 * 
 *     def __cinit__(self, *args, **kw):
 *         self.face = ""             # <<<<<<<<<<<<<<
 *         self.bold = False
 *         self.italic = False
 */
  __Pyx_INCREF(__pyx_kp_u__3);
  __Pyx_GIVEREF(__pyx_kp_u__3);
  __Pyx_GOTREF(__pyx_v_self->face);
  __Pyx_DECREF(__pyx_v_self->face);
  __pyx_v_self->face = __pyx_kp_u__3;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":652
 *     def __cinit__(self, *args, **kw):
 *         self.face = ""
 *         self.bold = False             # <<<<<<<<<<<<<<
 *         self.italic = False
 *         self.padding = 0
 */
  __pyx_v_self->bold = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":653
 *         self.face = ""
 *         self.bold = False
 *         self.italic = False             # <<<<<<<<<<<<<<
 *         self.padding = 0
 *         self.spacing = 0
 */
  __pyx_v_self->italic = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":654
 *         self.bold = False
 *         self.italic = False
 *         self.padding = 0             # <<<<<<<<<<<<<<
 *         self.spacing = 0
 *         self.outline = 0
 */
  __pyx_v_self->padding = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":655
 *         self.italic = False
 *         self.padding = 0
 *         self.spacing = 0             # <<<<<<<<<<<<<<
 *         self.outline = 0
 * 
 */
  __pyx_v_self->spacing = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":656
 *         self.padding = 0
 *         self.spacing = 0
 *         self.outline = 0             # <<<<<<<<<<<<<<
 * 
 *         self.line_height = 0
 */
  __pyx_v_self->outline = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":658
 *         self.outline = 0
 * 
 *         self.line_height = 0             # <<<<<<<<<<<<<<
 *         self.base = 0
 *         self.scale_w = 0
 */
  __pyx_v_self->line_height = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":659
 * 
 *         self.line_height = 0
 *         self.base = 0             # <<<<<<<<<<<<<<
 *         self.scale_w = 0
 *         self.scale_h = 0
 */
  __pyx_v_self->base = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":660
 *         self.line_height = 0
 *         self.base = 0
 *         self.scale_w = 0             # <<<<<<<<<<<<<<
 *         self.scale_h = 0
 * 
 */
  __pyx_v_self->scale_w = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":661
 *         self.base = 0
 *         self.scale_w = 0
 *         self.scale_h = 0             # <<<<<<<<<<<<<<
 * 
 *         self.characters = dict()
 */
  __pyx_v_self->scale_h = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":663
 *         self.scale_h = 0
 * 
 *         self.characters = dict()             # <<<<<<<<<<<<<<
 *         self.kernings = dict()
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->characters);
  __Pyx_DECREF(__pyx_v_self->characters);
  __pyx_v_self->characters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":664
 * 
 *         self.characters = dict()
 *         self.kernings = dict()             # <<<<<<<<<<<<<<
 * 
 *         self.image = NULL
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->kernings);
  __Pyx_DECREF(__pyx_v_self->kernings);
  __pyx_v_self->kernings = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":666
 *         self.kernings = dict()
 * 
 *         self.image = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, str image_file, object descriptor):
 */
  __pyx_v_self->image = NULL;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":650
 *     cdef SDL_Surface *image
 * 
 *     def __cinit__(self, *args, **kw):             # <<<<<<<<<<<<<<
 *         self.face = ""
 *         self.bold = False
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":668
 *         self.image = NULL
 * 
 *     def __init__(self, str image_file, object descriptor):             # <<<<<<<<<<<<<<
 * 
 *         # Load the image file
 */

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image_file = 0;
  PyObject *__pyx_v_descriptor = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image_file,&__pyx_n_s_descriptor,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image_file)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descriptor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(1, 668, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 668, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_image_file = ((PyObject*)values[0]);
    __pyx_v_descriptor = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 668, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_image_file), (&PyUnicode_Type), 1, "image_file", 1))) __PYX_ERR(1, 668, __pyx_L1_error)
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_2__init__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), __pyx_v_image_file, __pyx_v_descriptor);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_2__init__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_image_file, PyObject *__pyx_v_descriptor) {
  PyObject *__pyx_v_c_filename = 0;
  struct SDL_Surface *__pyx_v_image;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":671
 * 
 *         # Load the image file
 *         cdef bytes c_filename = image_file.encode('utf-8')             # <<<<<<<<<<<<<<
 *         cdef SDL_Surface *image = IMG_Load(c_filename)
 * 
 */
  if (unlikely(__pyx_v_image_file == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(1, 671, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsUTF8String(__pyx_v_image_file); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_c_filename = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":672
 *         # Load the image file
 *         cdef bytes c_filename = image_file.encode('utf-8')
 *         cdef SDL_Surface *image = IMG_Load(c_filename)             # <<<<<<<<<<<<<<
 * 
 *         if image == NULL:
 */
  if (unlikely(__pyx_v_c_filename == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 672, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_c_filename); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(1, 672, __pyx_L1_error)
  __pyx_v_image = IMG_Load(__pyx_t_2);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":674
 *         cdef SDL_Surface *image = IMG_Load(c_filename)
 * 
 *         if image == NULL:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Could not load bitmap font image file")
 * 
 */
  __pyx_t_3 = ((__pyx_v_image == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":675
 * 
 *         if image == NULL:
 *             raise BitmapFontException("Could not load bitmap font image file")             # <<<<<<<<<<<<<<
 * 
 *         self.image = image
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_Could_not_load_bitmap_font_image) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_Could_not_load_bitmap_font_image);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 675, __pyx_L1_error)

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":674
 *         cdef SDL_Surface *image = IMG_Load(c_filename)
 * 
 *         if image == NULL:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Could not load bitmap font image file")
 * 
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":677
 *             raise BitmapFontException("Could not load bitmap font image file")
 * 
 *         self.image = image             # <<<<<<<<<<<<<<
 *         self.scale_w = self.image.w
 *         self.scale_h = self.image.h
 */
  __pyx_v_self->image = __pyx_v_image;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":678
 * 
 *         self.image = image
 *         self.scale_w = self.image.w             # <<<<<<<<<<<<<<
 *         self.scale_h = self.image.h
 * 
 */
  __pyx_t_6 = __pyx_v_self->image->w;
  __pyx_v_self->scale_w = __pyx_t_6;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":679
 *         self.image = image
 *         self.scale_w = self.image.w
 *         self.scale_h = self.image.h             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(descriptor, list):
 */
  __pyx_t_6 = __pyx_v_self->image->h;
  __pyx_v_self->scale_h = __pyx_t_6;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":681
 *         self.scale_h = self.image.h
 * 
 *         if isinstance(descriptor, list):             # <<<<<<<<<<<<<<
 *             self._load_descriptor_list(descriptor)
 *         elif isinstance(descriptor, str):
 */
  __pyx_t_3 = PyList_Check(__pyx_v_descriptor); 
  __pyx_t_7 = (__pyx_t_3 != 0);
  if (__pyx_t_7) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":682
 * 
 *         if isinstance(descriptor, list):
 *             self._load_descriptor_list(descriptor)             # <<<<<<<<<<<<<<
 *         elif isinstance(descriptor, str):
 *             self._load_descriptor_file(descriptor)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_load_descriptor_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 682, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_descriptor) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_descriptor);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 682, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":681
 *         self.scale_h = self.image.h
 * 
 *         if isinstance(descriptor, list):             # <<<<<<<<<<<<<<
 *             self._load_descriptor_list(descriptor)
 *         elif isinstance(descriptor, str):
 */
    goto __pyx_L4;
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":683
 *         if isinstance(descriptor, list):
 *             self._load_descriptor_list(descriptor)
 *         elif isinstance(descriptor, str):             # <<<<<<<<<<<<<<
 *             self._load_descriptor_file(descriptor)
 *         else:
 */
  __pyx_t_7 = PyUnicode_Check(__pyx_v_descriptor); 
  __pyx_t_3 = (__pyx_t_7 != 0);
  if (likely(__pyx_t_3)) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":684
 *             self._load_descriptor_list(descriptor)
 *         elif isinstance(descriptor, str):
 *             self._load_descriptor_file(descriptor)             # <<<<<<<<<<<<<<
 *         else:
 *             raise BitmapFontException("Illegal value in bitmap font descriptor")
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_load_descriptor_file); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_descriptor) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_descriptor);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":683
 *         if isinstance(descriptor, list):
 *             self._load_descriptor_list(descriptor)
 *         elif isinstance(descriptor, str):             # <<<<<<<<<<<<<<
 *             self._load_descriptor_file(descriptor)
 *         else:
 */
    goto __pyx_L4;
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":686
 *             self._load_descriptor_file(descriptor)
 *         else:
 *             raise BitmapFontException("Illegal value in bitmap font descriptor")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 686, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_Illegal_value_in_bitmap_font_des) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_Illegal_value_in_bitmap_font_des);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 686, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 686, __pyx_L1_error)
  }
  __pyx_L4:;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":668
 *         self.image = NULL
 * 
 *     def __init__(self, str image_file, object descriptor):             # <<<<<<<<<<<<<<
 * 
 *         # Load the image file
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_c_filename);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":688
 *             raise BitmapFontException("Illegal value in bitmap font descriptor")
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.image != NULL:
 *             SDL_FreeSurface(self.image)
 */

/* Python wrapper */
static void __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_5__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4__dealloc__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4__dealloc__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":689
 * 
 *     def __dealloc__(self):
 *         if self.image != NULL:             # <<<<<<<<<<<<<<
 *             SDL_FreeSurface(self.image)
 *             self.image = NULL
 */
  __pyx_t_1 = ((__pyx_v_self->image != NULL) != 0);
  if (__pyx_t_1) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":690
 *     def __dealloc__(self):
 *         if self.image != NULL:
 *             SDL_FreeSurface(self.image)             # <<<<<<<<<<<<<<
 *             self.image = NULL
 * 
 */
    SDL_FreeSurface(__pyx_v_self->image);

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":691
 *         if self.image != NULL:
 *             SDL_FreeSurface(self.image)
 *             self.image = NULL             # <<<<<<<<<<<<<<
 * 
 *     def get_image(self):
 */
    __pyx_v_self->image = NULL;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":689
 * 
 *     def __dealloc__(self):
 *         if self.image != NULL:             # <<<<<<<<<<<<<<
 *             SDL_FreeSurface(self.image)
 *             self.image = NULL
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":688
 *             raise BitmapFontException("Illegal value in bitmap font descriptor")
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.image != NULL:
 *             SDL_FreeSurface(self.image)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":693
 *             self.image = NULL
 * 
 *     def get_image(self):             # <<<<<<<<<<<<<<
 *         image_capsule = pycapsule.PyCapsule_New(self.image, NULL, NULL)
 *         return image_capsule
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7get_image(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6get_image[] = "BitmapFont.get_image(self)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7get_image(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_image (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6get_image(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6get_image(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_v_image_capsule = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_image", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":694
 * 
 *     def get_image(self):
 *         image_capsule = pycapsule.PyCapsule_New(self.image, NULL, NULL)             # <<<<<<<<<<<<<<
 *         return image_capsule
 * 
 */
  __pyx_t_1 = PyCapsule_New(__pyx_v_self->image, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_image_capsule = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":695
 *     def get_image(self):
 *         image_capsule = pycapsule.PyCapsule_New(self.image, NULL, NULL)
 *         return image_capsule             # <<<<<<<<<<<<<<
 * 
 *     def get_characters(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_image_capsule);
  __pyx_r = __pyx_v_image_capsule;
  goto __pyx_L0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":693
 *             self.image = NULL
 * 
 *     def get_image(self):             # <<<<<<<<<<<<<<
 *         image_capsule = pycapsule.PyCapsule_New(self.image, NULL, NULL)
 *         return image_capsule
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.get_image", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_image_capsule);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":697
 *         return image_capsule
 * 
 *     def get_characters(self):             # <<<<<<<<<<<<<<
 *         return self.characters
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_9get_characters(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_8get_characters[] = "BitmapFont.get_characters(self)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_9get_characters(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_characters (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_8get_characters(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_8get_characters(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_characters", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":698
 * 
 *     def get_characters(self):
 *         return self.characters             # <<<<<<<<<<<<<<
 * 
 *     def get_kernings(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->characters);
  __pyx_r = __pyx_v_self->characters;
  goto __pyx_L0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":697
 *         return image_capsule
 * 
 *     def get_characters(self):             # <<<<<<<<<<<<<<
 *         return self.characters
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":700
 *         return self.characters
 * 
 *     def get_kernings(self):             # <<<<<<<<<<<<<<
 *         return self.kernings
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11get_kernings(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_10get_kernings[] = "BitmapFont.get_kernings(self)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11get_kernings(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_kernings (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_10get_kernings(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_10get_kernings(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_kernings", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":701
 * 
 *     def get_kernings(self):
 *         return self.kernings             # <<<<<<<<<<<<<<
 * 
 *     def _load_descriptor_list(self, list descriptor_list):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->kernings);
  __pyx_r = __pyx_v_self->kernings;
  goto __pyx_L0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":700
 *         return self.characters
 * 
 *     def get_kernings(self):             # <<<<<<<<<<<<<<
 *         return self.kernings
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":703
 *         return self.kernings
 * 
 *     def _load_descriptor_list(self, list descriptor_list):             # <<<<<<<<<<<<<<
 *         """Load the descriptor from the supplied list."""
 *         cdef int x = 0
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_13_load_descriptor_list(PyObject *__pyx_v_self, PyObject *__pyx_v_descriptor_list); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_12_load_descriptor_list[] = "BitmapFont._load_descriptor_list(self, list descriptor_list)\nLoad the descriptor from the supplied list.";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_13_load_descriptor_list(PyObject *__pyx_v_self, PyObject *__pyx_v_descriptor_list) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_load_descriptor_list (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_descriptor_list), (&PyList_Type), 1, "descriptor_list", 1))) __PYX_ERR(1, 703, __pyx_L1_error)
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_12_load_descriptor_list(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject*)__pyx_v_descriptor_list));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_12_load_descriptor_list(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_descriptor_list) {
  int __pyx_v_x;
  int __pyx_v_y;
  int __pyx_v_row_height;
  int __pyx_v_char_width;
  int __pyx_v_char_id;
  PyObject *__pyx_v_row = NULL;
  PyObject *__pyx_v_text_char = NULL;
  struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_character = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_load_descriptor_list", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":705
 *     def _load_descriptor_list(self, list descriptor_list):
 *         """Load the descriptor from the supplied list."""
 *         cdef int x = 0             # <<<<<<<<<<<<<<
 *         cdef int y = 0
 *         cdef int row_height = int(self.scale_h / len(descriptor_list))
 */
  __pyx_v_x = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":706
 *         """Load the descriptor from the supplied list."""
 *         cdef int x = 0
 *         cdef int y = 0             # <<<<<<<<<<<<<<
 *         cdef int row_height = int(self.scale_h / len(descriptor_list))
 *         cdef int char_width
 */
  __pyx_v_y = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":707
 *         cdef int x = 0
 *         cdef int y = 0
 *         cdef int row_height = int(self.scale_h / len(descriptor_list))             # <<<<<<<<<<<<<<
 *         cdef int char_width
 *         cdef int char_id
 */
  if (unlikely(__pyx_v_descriptor_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 707, __pyx_L1_error)
  }
  __pyx_t_1 = PyList_GET_SIZE(__pyx_v_descriptor_list); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 707, __pyx_L1_error)
  if (unlikely(__pyx_t_1 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(1, 707, __pyx_L1_error)
  }
  __pyx_v_row_height = ((int)(((double)__pyx_v_self->scale_h) / ((double)__pyx_t_1)));

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":710
 *         cdef int char_width
 *         cdef int char_id
 *         self.line_height = row_height             # <<<<<<<<<<<<<<
 *         self.base = row_height
 * 
 */
  __pyx_v_self->line_height = __pyx_v_row_height;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":711
 *         cdef int char_id
 *         self.line_height = row_height
 *         self.base = row_height             # <<<<<<<<<<<<<<
 * 
 *         # Loop over all the rows in the descriptor list
 */
  __pyx_v_self->base = __pyx_v_row_height;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":714
 * 
 *         # Loop over all the rows in the descriptor list
 *         for row in descriptor_list:             # <<<<<<<<<<<<<<
 *             char_width = int(self.scale_w / len(row))
 *             x = 0
 */
  if (unlikely(__pyx_v_descriptor_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 714, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_descriptor_list; __Pyx_INCREF(__pyx_t_2); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(1, 714, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_row, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":715
 *         # Loop over all the rows in the descriptor list
 *         for row in descriptor_list:
 *             char_width = int(self.scale_w / len(row))             # <<<<<<<<<<<<<<
 *             x = 0
 * 
 */
    __pyx_t_4 = PyObject_Length(__pyx_v_row); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 715, __pyx_L1_error)
    if (unlikely(__pyx_t_4 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(1, 715, __pyx_L1_error)
    }
    __pyx_v_char_width = ((int)(((double)__pyx_v_self->scale_w) / ((double)__pyx_t_4)));

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":716
 *         for row in descriptor_list:
 *             char_width = int(self.scale_w / len(row))
 *             x = 0             # <<<<<<<<<<<<<<
 * 
 *             # Loop over all characters in the row
 */
    __pyx_v_x = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":719
 * 
 *             # Loop over all characters in the row
 *             for text_char in row:             # <<<<<<<<<<<<<<
 *                 char_id = ord(text_char)
 * 
 */
    if (likely(PyList_CheckExact(__pyx_v_row)) || PyTuple_CheckExact(__pyx_v_row)) {
      __pyx_t_3 = __pyx_v_row; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_row); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 719, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 719, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(1, 719, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 719, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(1, 719, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 719, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(1, 719, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_text_char, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":720
 *             # Loop over all characters in the row
 *             for text_char in row:
 *                 char_id = ord(text_char)             # <<<<<<<<<<<<<<
 * 
 *                 # Do not add duplicate character definitions (only use the first instance)
 */
      __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_v_text_char); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(1, 720, __pyx_L1_error)
      __pyx_v_char_id = __pyx_t_7;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":723
 * 
 *                 # Do not add duplicate character definitions (only use the first instance)
 *                 if char_id not in self.characters:             # <<<<<<<<<<<<<<
 * 
 *                     # Create character definition
 */
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_char_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 723, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__pyx_v_self->characters == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(1, 723, __pyx_L1_error)
      }
      __pyx_t_8 = (__Pyx_PyDict_ContainsTF(__pyx_t_6, __pyx_v_self->characters, Py_NE)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(1, 723, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_9 = (__pyx_t_8 != 0);
      if (__pyx_t_9) {

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":726
 * 
 *                     # Create character definition
 *                     character = BitmapFontCharacter()             # <<<<<<<<<<<<<<
 *                     character.id = char_id
 *                     character.rect.x = x
 */
        __pyx_t_6 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_v_character, ((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":727
 *                     # Create character definition
 *                     character = BitmapFontCharacter()
 *                     character.id = char_id             # <<<<<<<<<<<<<<
 *                     character.rect.x = x
 *                     character.rect.y = y
 */
        __pyx_v_character->id = __pyx_v_char_id;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":728
 *                     character = BitmapFontCharacter()
 *                     character.id = char_id
 *                     character.rect.x = x             # <<<<<<<<<<<<<<
 *                     character.rect.y = y
 *                     character.rect.w = char_width
 */
        __pyx_v_character->rect.x = __pyx_v_x;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":729
 *                     character.id = char_id
 *                     character.rect.x = x
 *                     character.rect.y = y             # <<<<<<<<<<<<<<
 *                     character.rect.w = char_width
 *                     character.rect.h = row_height
 */
        __pyx_v_character->rect.y = __pyx_v_y;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":730
 *                     character.rect.x = x
 *                     character.rect.y = y
 *                     character.rect.w = char_width             # <<<<<<<<<<<<<<
 *                     character.rect.h = row_height
 *                     character.xadvance = char_width
 */
        __pyx_v_character->rect.w = __pyx_v_char_width;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":731
 *                     character.rect.y = y
 *                     character.rect.w = char_width
 *                     character.rect.h = row_height             # <<<<<<<<<<<<<<
 *                     character.xadvance = char_width
 * 
 */
        __pyx_v_character->rect.h = __pyx_v_row_height;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":732
 *                     character.rect.w = char_width
 *                     character.rect.h = row_height
 *                     character.xadvance = char_width             # <<<<<<<<<<<<<<
 * 
 *                     self.characters[character.id] = character
 */
        __pyx_v_character->xadvance = __pyx_v_char_width;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":734
 *                     character.xadvance = char_width
 * 
 *                     self.characters[character.id] = character             # <<<<<<<<<<<<<<
 *                     x += char_width
 * 
 */
        if (unlikely(__pyx_v_self->characters == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(1, 734, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_character->id); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 734, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(PyDict_SetItem(__pyx_v_self->characters, __pyx_t_6, ((PyObject *)__pyx_v_character)) < 0)) __PYX_ERR(1, 734, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":735
 * 
 *                     self.characters[character.id] = character
 *                     x += char_width             # <<<<<<<<<<<<<<
 * 
 *             y += row_height
 */
        __pyx_v_x = (__pyx_v_x + __pyx_v_char_width);

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":723
 * 
 *                 # Do not add duplicate character definitions (only use the first instance)
 *                 if char_id not in self.characters:             # <<<<<<<<<<<<<<
 * 
 *                     # Create character definition
 */
      }

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":719
 * 
 *             # Loop over all characters in the row
 *             for text_char in row:             # <<<<<<<<<<<<<<
 *                 char_id = ord(text_char)
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":737
 *                     x += char_width
 * 
 *             y += row_height             # <<<<<<<<<<<<<<
 * 
 *     def _load_descriptor_file(self, str descriptor_file):
 */
    __pyx_v_y = (__pyx_v_y + __pyx_v_row_height);

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":714
 * 
 *         # Loop over all the rows in the descriptor list
 *         for row in descriptor_list:             # <<<<<<<<<<<<<<
 *             char_width = int(self.scale_w / len(row))
 *             x = 0
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":703
 *         return self.kernings
 * 
 *     def _load_descriptor_list(self, list descriptor_list):             # <<<<<<<<<<<<<<
 *         """Load the descriptor from the supplied list."""
 *         cdef int x = 0
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont._load_descriptor_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_row);
  __Pyx_XDECREF(__pyx_v_text_char);
  __Pyx_XDECREF((PyObject *)__pyx_v_character);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":739
 *             y += row_height
 * 
 *     def _load_descriptor_file(self, str descriptor_file):             # <<<<<<<<<<<<<<
 *         """Load the descriptor from the specified file.  Both XML and text
 *         formats are supported (not binary currently).
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_15_load_descriptor_file(PyObject *__pyx_v_self, PyObject *__pyx_v_descriptor_file); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_14_load_descriptor_file[] = "BitmapFont._load_descriptor_file(self, unicode descriptor_file)\nLoad the descriptor from the specified file.  Both XML and text\n        formats are supported (not binary currently).\n\n        The standard bitmap font descriptor file format can be found at:\n        http://www.angelcode.com/products/bmfont/doc/file_format.html\n        ";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_15_load_descriptor_file(PyObject *__pyx_v_self, PyObject *__pyx_v_descriptor_file) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_load_descriptor_file (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_descriptor_file), (&PyUnicode_Type), 1, "descriptor_file", 1))) __PYX_ERR(1, 739, __pyx_L1_error)
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_14_load_descriptor_file(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject*)__pyx_v_descriptor_file));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_14_load_descriptor_file(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_descriptor_file) {
  int __pyx_v_char_id;
  int __pyx_v_first;
  int __pyx_v_second;
  PyObject *__pyx_v_xml_tree = NULL;
  PyObject *__pyx_v_text_file = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_v_m = NULL;
  struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_character = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_load_descriptor_file", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":750
 *         cdef int second
 * 
 *         if not path.isfile(descriptor_file):             # <<<<<<<<<<<<<<
 *             raise BitmapFontException('Could not locate the bitmap font descriptor file ' +
 *                                       descriptor_file)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_isfile); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_descriptor_file) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_descriptor_file);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":751
 * 
 *         if not path.isfile(descriptor_file):
 *             raise BitmapFontException('Could not locate the bitmap font descriptor file ' +             # <<<<<<<<<<<<<<
 *                                       descriptor_file)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":752
 *         if not path.isfile(descriptor_file):
 *             raise BitmapFontException('Could not locate the bitmap font descriptor file ' +
 *                                       descriptor_file)             # <<<<<<<<<<<<<<
 * 
 *         # Attempt to parse the file as an XML file
 */
    __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_kp_u_Could_not_locate_the_bitmap_font, __pyx_v_descriptor_file); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 751, __pyx_L1_error)

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":750
 *         cdef int second
 * 
 *         if not path.isfile(descriptor_file):             # <<<<<<<<<<<<<<
 *             raise BitmapFontException('Could not locate the bitmap font descriptor file ' +
 *                                       descriptor_file)
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":755
 * 
 *         # Attempt to parse the file as an XML file
 *         try:             # <<<<<<<<<<<<<<
 *             xml_tree = ElementTree(file=descriptor_file)
 *             self._load_descriptor_xml(xml_tree)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":756
 *         # Attempt to parse the file as an XML file
 *         try:
 *             xml_tree = ElementTree(file=descriptor_file)             # <<<<<<<<<<<<<<
 *             self._load_descriptor_xml(xml_tree)
 *         except ParseError:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ElementTree); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 756, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 756, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_file, __pyx_v_descriptor_file) < 0) __PYX_ERR(1, 756, __pyx_L4_error)
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 756, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_xml_tree = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":757
 *         try:
 *             xml_tree = ElementTree(file=descriptor_file)
 *             self._load_descriptor_xml(xml_tree)             # <<<<<<<<<<<<<<
 *         except ParseError:
 *             pass
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_load_descriptor_xml); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 757, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_v_xml_tree) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_xml_tree);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 757, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":755
 * 
 *         # Attempt to parse the file as an XML file
 *         try:             # <<<<<<<<<<<<<<
 *             xml_tree = ElementTree(file=descriptor_file)
 *             self._load_descriptor_xml(xml_tree)
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":758
 *             xml_tree = ElementTree(file=descriptor_file)
 *             self._load_descriptor_xml(xml_tree)
 *         except ParseError:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_3, &__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ParseError); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 758, __pyx_L6_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_2, __pyx_t_3, __pyx_t_1);
    __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_1 = 0;
    if (__pyx_t_10) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":755
 * 
 *         # Attempt to parse the file as an XML file
 *         try:             # <<<<<<<<<<<<<<
 *             xml_tree = ElementTree(file=descriptor_file)
 *             self._load_descriptor_xml(xml_tree)
 */
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    __pyx_L9_try_end:;
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":764
 * 
 *         # Open the descriptor file
 *         with open(descriptor_file) as text_file:             # <<<<<<<<<<<<<<
 * 
 *             # Loop over all the rows in the descriptor file
 */
  /*with:*/ {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_open, __pyx_v_descriptor_file); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_2 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 764, __pyx_L10_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 764, __pyx_L10_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {
          __pyx_v_text_file = __pyx_t_2;
          __pyx_t_2 = 0;

          /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":767
 * 
 *             # Loop over all the rows in the descriptor file
 *             for line in text_file:             # <<<<<<<<<<<<<<
 *                 if line.startswith("info"):
 *                     pass
 */
          if (likely(PyList_CheckExact(__pyx_v_text_file)) || PyTuple_CheckExact(__pyx_v_text_file)) {
            __pyx_t_2 = __pyx_v_text_file; __Pyx_INCREF(__pyx_t_2); __pyx_t_12 = 0;
            __pyx_t_13 = NULL;
          } else {
            __pyx_t_12 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_text_file); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 767, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_13 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 767, __pyx_L14_error)
          }
          for (;;) {
            if (likely(!__pyx_t_13)) {
              if (likely(PyList_CheckExact(__pyx_t_2))) {
                if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_2)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(1, 767, __pyx_L14_error)
                #else
                __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 767, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_1);
                #endif
              } else {
                if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(1, 767, __pyx_L14_error)
                #else
                __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 767, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_1);
                #endif
              }
            } else {
              __pyx_t_1 = __pyx_t_13(__pyx_t_2);
              if (unlikely(!__pyx_t_1)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(1, 767, __pyx_L14_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_1);
            }
            __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":768
 *             # Loop over all the rows in the descriptor file
 *             for line in text_file:
 *                 if line.startswith("info"):             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_startswith); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 768, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_6 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_n_u_info) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_info);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 768, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 768, __pyx_L14_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_5) {
              goto __pyx_L22;
            }

            /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":771
 *                     pass
 * 
 *                 elif line.startswith("common"):             # <<<<<<<<<<<<<<
 *                     m = re.search(r"lineHeight=([0-9]{1,5})", line, flags=re.IGNORECASE)
 *                     if m:
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_startswith); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 771, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_6 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_n_u_common) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_common);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 771, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 771, __pyx_L14_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_5) {

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":772
 * 
 *                 elif line.startswith("common"):
 *                     m = re.search(r"lineHeight=([0-9]{1,5})", line, flags=re.IGNORECASE)             # <<<<<<<<<<<<<<
 *                     if m:
 *                         self.line_height = int(m.group(1))
 */
              __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 772, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_search); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 772, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 772, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_INCREF(__pyx_kp_u_lineHeight_0_9_1_5);
              __Pyx_GIVEREF(__pyx_kp_u_lineHeight_0_9_1_5);
              PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_lineHeight_0_9_1_5);
              __Pyx_INCREF(__pyx_v_line);
              __Pyx_GIVEREF(__pyx_v_line);
              PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_line);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 772, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_re); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 772, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_IGNORECASE); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 772, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_flags, __pyx_t_15) < 0) __PYX_ERR(1, 772, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 772, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_15);
              __pyx_t_15 = 0;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":773
 *                 elif line.startswith("common"):
 *                     m = re.search(r"lineHeight=([0-9]{1,5})", line, flags=re.IGNORECASE)
 *                     if m:             # <<<<<<<<<<<<<<
 *                         self.line_height = int(m.group(1))
 *                     else:
 */
              __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_m); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 773, __pyx_L14_error)
              if (likely(__pyx_t_5)) {

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":774
 *                     m = re.search(r"lineHeight=([0-9]{1,5})", line, flags=re.IGNORECASE)
 *                     if m:
 *                         self.line_height = int(m.group(1))             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 */
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 774, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_15 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_int_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_int_1);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 774, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_15);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_15); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 774, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 774, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_v_self->line_height = __pyx_t_10;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":773
 *                 elif line.startswith("common"):
 *                     m = re.search(r"lineHeight=([0-9]{1,5})", line, flags=re.IGNORECASE)
 *                     if m:             # <<<<<<<<<<<<<<
 *                         self.line_height = int(m.group(1))
 *                     else:
 */
                goto __pyx_L23;
              }

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":776
 *                         self.line_height = int(m.group(1))
 *                     else:
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")             # <<<<<<<<<<<<<<
 * 
 *                     m = re.search(r"base=([0-9]{1,5})", line, flags=re.IGNORECASE)
 */
              /*else*/ {
                __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 776, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_15);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_15, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_1, __pyx_kp_u_Bitmap_font_descriptor_file_inva) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_kp_u_Bitmap_font_descriptor_file_inva);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 776, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_Raise(__pyx_t_6, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __PYX_ERR(1, 776, __pyx_L14_error)
              }
              __pyx_L23:;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":778
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 * 
 *                     m = re.search(r"base=([0-9]{1,5})", line, flags=re.IGNORECASE)             # <<<<<<<<<<<<<<
 *                     if m:
 *                         self.base = int(m.group(1))
 */
              __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_re); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 778, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_search); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 778, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 778, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_INCREF(__pyx_kp_u_base_0_9_1_5);
              __Pyx_GIVEREF(__pyx_kp_u_base_0_9_1_5);
              PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_u_base_0_9_1_5);
              __Pyx_INCREF(__pyx_v_line);
              __Pyx_GIVEREF(__pyx_v_line);
              PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_line);
              __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 778, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_re); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 778, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_IGNORECASE); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 778, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_flags, __pyx_t_14) < 0) __PYX_ERR(1, 778, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 778, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF_SET(__pyx_v_m, __pyx_t_14);
              __pyx_t_14 = 0;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":779
 * 
 *                     m = re.search(r"base=([0-9]{1,5})", line, flags=re.IGNORECASE)
 *                     if m:             # <<<<<<<<<<<<<<
 *                         self.base = int(m.group(1))
 *                     else:
 */
              __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_m); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 779, __pyx_L14_error)
              if (likely(__pyx_t_5)) {

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":780
 *                     m = re.search(r"base=([0-9]{1,5})", line, flags=re.IGNORECASE)
 *                     if m:
 *                         self.base = int(m.group(1))             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 */
                __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 780, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_6 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_1, function);
                  }
                }
                __pyx_t_14 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_int_1) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_int_1);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 780, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_14);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 780, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 780, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __pyx_v_self->base = __pyx_t_10;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":779
 * 
 *                     m = re.search(r"base=([0-9]{1,5})", line, flags=re.IGNORECASE)
 *                     if m:             # <<<<<<<<<<<<<<
 *                         self.base = int(m.group(1))
 *                     else:
 */
                goto __pyx_L24;
              }

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":782
 *                         self.base = int(m.group(1))
 *                     else:
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")             # <<<<<<<<<<<<<<
 * 
 *                 elif line.startswith("chars"):
 */
              /*else*/ {
                __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 782, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_14);
                __pyx_t_6 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_14);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_14, function);
                  }
                }
                __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_6, __pyx_kp_u_Bitmap_font_descriptor_file_inva) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_kp_u_Bitmap_font_descriptor_file_inva);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 782, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_Raise(__pyx_t_1, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __PYX_ERR(1, 782, __pyx_L14_error)
              }
              __pyx_L24:;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":771
 *                     pass
 * 
 *                 elif line.startswith("common"):             # <<<<<<<<<<<<<<
 *                     m = re.search(r"lineHeight=([0-9]{1,5})", line, flags=re.IGNORECASE)
 *                     if m:
 */
              goto __pyx_L22;
            }

            /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":784
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 * 
 *                 elif line.startswith("chars"):             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
            __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_startswith); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 784, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_6 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_14);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_14, function);
              }
            }
            __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_6, __pyx_n_u_chars) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_n_u_chars);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 784, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 784, __pyx_L14_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_5) {
              goto __pyx_L22;
            }

            /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":787
 *                     pass
 * 
 *                 elif line.startswith("char"):             # <<<<<<<<<<<<<<
 *                     m = re.search(r"char\s+id=(?P<id>[0-9]{1,4})\s+x=(?P<x>[0-9]{1,4})\s+y=(?P<y>[0-9]{1,4})"
 *                                   r"\s+width=(?P<width>[0-9]{1,3})\s+height=(?P<height>[0-9]{1,3})"
 */
            __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_startswith); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 787, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_6 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
              __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_14);
              if (likely(__pyx_t_6)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_14, function);
              }
            }
            __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_6, __pyx_n_u_char) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_n_u_char);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 787, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 787, __pyx_L14_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_5) {

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":788
 * 
 *                 elif line.startswith("char"):
 *                     m = re.search(r"char\s+id=(?P<id>[0-9]{1,4})\s+x=(?P<x>[0-9]{1,4})\s+y=(?P<y>[0-9]{1,4})"             # <<<<<<<<<<<<<<
 *                                   r"\s+width=(?P<width>[0-9]{1,3})\s+height=(?P<height>[0-9]{1,3})"
 *                                   r"\s+xoffset=(?P<xoffset>-?[0-9]{1,3})\s+yoffset=(?P<yoffset>-?[0-9]{1,3})"
 */
              __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 788, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_search); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 788, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":792
 *                                   r"\s+xoffset=(?P<xoffset>-?[0-9]{1,3})\s+yoffset=(?P<yoffset>-?[0-9]{1,3})"
 *                                   r"\s+xadvance=(?P<xadvance>[0-9]{1,3})\s+page=(?P<page>[0-9]{1,2})"
 *                                   r"\s+chnl=(?P<chnl>[0-9]{1,3})", line, flags=re.IGNORECASE)             # <<<<<<<<<<<<<<
 *                     if not m:
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 */
              __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 788, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_INCREF(__pyx_kp_u_char_s_id_P_id_0_9_1_4_s_x_P_x_0);
              __Pyx_GIVEREF(__pyx_kp_u_char_s_id_P_id_0_9_1_4_s_x_P_x_0);
              PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_char_s_id_P_id_0_9_1_4_s_x_P_x_0);
              __Pyx_INCREF(__pyx_v_line);
              __Pyx_GIVEREF(__pyx_v_line);
              PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_line);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 792, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_re); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 792, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_IGNORECASE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 792, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_flags, __pyx_t_3) < 0) __PYX_ERR(1, 792, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":788
 * 
 *                 elif line.startswith("char"):
 *                     m = re.search(r"char\s+id=(?P<id>[0-9]{1,4})\s+x=(?P<x>[0-9]{1,4})\s+y=(?P<y>[0-9]{1,4})"             # <<<<<<<<<<<<<<
 *                                   r"\s+width=(?P<width>[0-9]{1,3})\s+height=(?P<height>[0-9]{1,3})"
 *                                   r"\s+xoffset=(?P<xoffset>-?[0-9]{1,3})\s+yoffset=(?P<yoffset>-?[0-9]{1,3})"
 */
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 788, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_3);
              __pyx_t_3 = 0;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":793
 *                                   r"\s+xadvance=(?P<xadvance>[0-9]{1,3})\s+page=(?P<page>[0-9]{1,2})"
 *                                   r"\s+chnl=(?P<chnl>[0-9]{1,3})", line, flags=re.IGNORECASE)
 *                     if not m:             # <<<<<<<<<<<<<<
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 * 
 */
              __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_m); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 793, __pyx_L14_error)
              __pyx_t_4 = ((!__pyx_t_5) != 0);
              if (unlikely(__pyx_t_4)) {

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":794
 *                                   r"\s+chnl=(?P<chnl>[0-9]{1,3})", line, flags=re.IGNORECASE)
 *                     if not m:
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")             # <<<<<<<<<<<<<<
 * 
 *                     char_id = int(m.group("id"))
 */
                __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 794, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_kp_u_Bitmap_font_descriptor_file_inva) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u_Bitmap_font_descriptor_file_inva);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 794, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_Raise(__pyx_t_3, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __PYX_ERR(1, 794, __pyx_L14_error)

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":793
 *                                   r"\s+xadvance=(?P<xadvance>[0-9]{1,3})\s+page=(?P<page>[0-9]{1,2})"
 *                                   r"\s+chnl=(?P<chnl>[0-9]{1,3})", line, flags=re.IGNORECASE)
 *                     if not m:             # <<<<<<<<<<<<<<
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 * 
 */
              }

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":796
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 * 
 *                     char_id = int(m.group("id"))             # <<<<<<<<<<<<<<
 *                     if char_id > 256:
 *                         continue
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 796, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_1 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_1)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_1);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_id) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_id);
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 796, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 796, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 796, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_v_char_id = __pyx_t_10;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":797
 * 
 *                     char_id = int(m.group("id"))
 *                     if char_id > 256:             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
              __pyx_t_4 = ((__pyx_v_char_id > 0x100) != 0);
              if (__pyx_t_4) {

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":798
 *                     char_id = int(m.group("id"))
 *                     if char_id > 256:
 *                         continue             # <<<<<<<<<<<<<<
 * 
 *                     # Do not add duplicate character definitions (only use the first instance)
 */
                goto __pyx_L20_continue;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":797
 * 
 *                     char_id = int(m.group("id"))
 *                     if char_id > 256:             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
              }

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":801
 * 
 *                     # Do not add duplicate character definitions (only use the first instance)
 *                     if char_id not in self.characters:             # <<<<<<<<<<<<<<
 * 
 *                         # Create character definition
 */
              __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_char_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 801, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (unlikely(__pyx_v_self->characters == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                __PYX_ERR(1, 801, __pyx_L14_error)
              }
              __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_t_6, __pyx_v_self->characters, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 801, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_5 = (__pyx_t_4 != 0);
              if (__pyx_t_5) {

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":804
 * 
 *                         # Create character definition
 *                         character = BitmapFontCharacter()             # <<<<<<<<<<<<<<
 *                         character.id = char_id
 *                         character.rect.x = int(m.group("x"))
 */
                __pyx_t_6 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 804, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_XDECREF_SET(__pyx_v_character, ((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_t_6));
                __pyx_t_6 = 0;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":805
 *                         # Create character definition
 *                         character = BitmapFontCharacter()
 *                         character.id = char_id             # <<<<<<<<<<<<<<
 *                         character.rect.x = int(m.group("x"))
 *                         character.rect.y = int(m.group("y"))
 */
                __pyx_v_character->id = __pyx_v_char_id;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":806
 *                         character = BitmapFontCharacter()
 *                         character.id = char_id
 *                         character.rect.x = int(m.group("x"))             # <<<<<<<<<<<<<<
 *                         character.rect.y = int(m.group("y"))
 *                         character.rect.w = int(m.group("width"))
 */
                __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 806, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_3, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_n_u_x) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_x);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 806, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 806, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 806, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_v_character->rect.x = __pyx_t_10;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":807
 *                         character.id = char_id
 *                         character.rect.x = int(m.group("x"))
 *                         character.rect.y = int(m.group("y"))             # <<<<<<<<<<<<<<
 *                         character.rect.w = int(m.group("width"))
 *                         character.rect.h = int(m.group("height"))
 */
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 807, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_y) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_y);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 807, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 807, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 807, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_v_character->rect.y = __pyx_t_10;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":808
 *                         character.rect.x = int(m.group("x"))
 *                         character.rect.y = int(m.group("y"))
 *                         character.rect.w = int(m.group("width"))             # <<<<<<<<<<<<<<
 *                         character.rect.h = int(m.group("height"))
 *                         character.xoffset = int(m.group("xoffset"))
 */
                __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 808, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_3, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_n_u_width) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_width);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 808, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 808, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 808, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_v_character->rect.w = __pyx_t_10;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":809
 *                         character.rect.y = int(m.group("y"))
 *                         character.rect.w = int(m.group("width"))
 *                         character.rect.h = int(m.group("height"))             # <<<<<<<<<<<<<<
 *                         character.xoffset = int(m.group("xoffset"))
 *                         character.yoffset = int(m.group("yoffset"))
 */
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 809, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_height) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_height);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 809, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 809, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 809, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_v_character->rect.h = __pyx_t_10;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":810
 *                         character.rect.w = int(m.group("width"))
 *                         character.rect.h = int(m.group("height"))
 *                         character.xoffset = int(m.group("xoffset"))             # <<<<<<<<<<<<<<
 *                         character.yoffset = int(m.group("yoffset"))
 *                         character.xadvance = int(m.group("xadvance"))
 */
                __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 810, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_3, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_n_u_xoffset) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_xoffset);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 810, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 810, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 810, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_v_character->xoffset = __pyx_t_10;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":811
 *                         character.rect.h = int(m.group("height"))
 *                         character.xoffset = int(m.group("xoffset"))
 *                         character.yoffset = int(m.group("yoffset"))             # <<<<<<<<<<<<<<
 *                         character.xadvance = int(m.group("xadvance"))
 * 
 */
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 811, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_yoffset) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_yoffset);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 811, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 811, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 811, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_v_character->yoffset = __pyx_t_10;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":812
 *                         character.xoffset = int(m.group("xoffset"))
 *                         character.yoffset = int(m.group("yoffset"))
 *                         character.xadvance = int(m.group("xadvance"))             # <<<<<<<<<<<<<<
 * 
 *                         self.characters[character.id] = character
 */
                __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 812, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_3, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_n_u_xadvance) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_xadvance);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 812, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 812, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 812, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __pyx_v_character->xadvance = __pyx_t_10;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":814
 *                         character.xadvance = int(m.group("xadvance"))
 * 
 *                         self.characters[character.id] = character             # <<<<<<<<<<<<<<
 * 
 *                 elif line.startswith("kernings"):
 */
                if (unlikely(__pyx_v_self->characters == Py_None)) {
                  PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                  __PYX_ERR(1, 814, __pyx_L14_error)
                }
                __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_character->id); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 814, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                if (unlikely(PyDict_SetItem(__pyx_v_self->characters, __pyx_t_3, ((PyObject *)__pyx_v_character)) < 0)) __PYX_ERR(1, 814, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":801
 * 
 *                     # Do not add duplicate character definitions (only use the first instance)
 *                     if char_id not in self.characters:             # <<<<<<<<<<<<<<
 * 
 *                         # Create character definition
 */
              }

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":787
 *                     pass
 * 
 *                 elif line.startswith("char"):             # <<<<<<<<<<<<<<
 *                     m = re.search(r"char\s+id=(?P<id>[0-9]{1,4})\s+x=(?P<x>[0-9]{1,4})\s+y=(?P<y>[0-9]{1,4})"
 *                                   r"\s+width=(?P<width>[0-9]{1,3})\s+height=(?P<height>[0-9]{1,3})"
 */
              goto __pyx_L22;
            }

            /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":816
 *                         self.characters[character.id] = character
 * 
 *                 elif line.startswith("kernings"):             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_startswith); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 816, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
              }
            }
            __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_kernings) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_kernings);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 816, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 816, __pyx_L14_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (__pyx_t_5) {
              goto __pyx_L22;
            }

            /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":819
 *                     pass
 * 
 *                 elif line.startswith("kerning"):             # <<<<<<<<<<<<<<
 *                     m = re.search(r"kerning\s+first=(?P<first>[0-9]{1,3})\s+second=(?P<second>[0-9]{1,3})"
 *                                   r"\s+amount=(?P<amount>-?[0-9]{1,3})", line, flags=re.IGNORECASE)
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_startswith); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 819, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
              }
            }
            __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_kerning) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_kerning);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 819, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 819, __pyx_L14_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (__pyx_t_5) {

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":820
 * 
 *                 elif line.startswith("kerning"):
 *                     m = re.search(r"kerning\s+first=(?P<first>[0-9]{1,3})\s+second=(?P<second>[0-9]{1,3})"             # <<<<<<<<<<<<<<
 *                                   r"\s+amount=(?P<amount>-?[0-9]{1,3})", line, flags=re.IGNORECASE)
 *                     if not m:
 */
              __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_re); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 820, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_search); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 820, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":821
 *                 elif line.startswith("kerning"):
 *                     m = re.search(r"kerning\s+first=(?P<first>[0-9]{1,3})\s+second=(?P<second>[0-9]{1,3})"
 *                                   r"\s+amount=(?P<amount>-?[0-9]{1,3})", line, flags=re.IGNORECASE)             # <<<<<<<<<<<<<<
 *                     if not m:
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 */
              __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 820, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_INCREF(__pyx_kp_u_kerning_s_first_P_first_0_9_1_3);
              __Pyx_GIVEREF(__pyx_kp_u_kerning_s_first_P_first_0_9_1_3);
              PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_kerning_s_first_P_first_0_9_1_3);
              __Pyx_INCREF(__pyx_v_line);
              __Pyx_GIVEREF(__pyx_v_line);
              PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_line);
              __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 821, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_re); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 821, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_IGNORECASE); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 821, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_flags, __pyx_t_15) < 0) __PYX_ERR(1, 821, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":820
 * 
 *                 elif line.startswith("kerning"):
 *                     m = re.search(r"kerning\s+first=(?P<first>[0-9]{1,3})\s+second=(?P<second>[0-9]{1,3})"             # <<<<<<<<<<<<<<
 *                                   r"\s+amount=(?P<amount>-?[0-9]{1,3})", line, flags=re.IGNORECASE)
 *                     if not m:
 */
              __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 820, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_15);
              __pyx_t_15 = 0;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":822
 *                     m = re.search(r"kerning\s+first=(?P<first>[0-9]{1,3})\s+second=(?P<second>[0-9]{1,3})"
 *                                   r"\s+amount=(?P<amount>-?[0-9]{1,3})", line, flags=re.IGNORECASE)
 *                     if not m:             # <<<<<<<<<<<<<<
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 * 
 */
              __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_m); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 822, __pyx_L14_error)
              __pyx_t_4 = ((!__pyx_t_5) != 0);
              if (unlikely(__pyx_t_4)) {

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":823
 *                                   r"\s+amount=(?P<amount>-?[0-9]{1,3})", line, flags=re.IGNORECASE)
 *                     if not m:
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")             # <<<<<<<<<<<<<<
 * 
 *                     first = int(m.group("first"))
 */
                __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 823, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_3 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
                  __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
                  if (likely(__pyx_t_3)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                    __Pyx_INCREF(__pyx_t_3);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_1, function);
                  }
                }
                __pyx_t_15 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_kp_u_Bitmap_font_descriptor_file_inva) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_u_Bitmap_font_descriptor_file_inva);
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 823, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_15);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_Raise(__pyx_t_15, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __PYX_ERR(1, 823, __pyx_L14_error)

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":822
 *                     m = re.search(r"kerning\s+first=(?P<first>[0-9]{1,3})\s+second=(?P<second>[0-9]{1,3})"
 *                                   r"\s+amount=(?P<amount>-?[0-9]{1,3})", line, flags=re.IGNORECASE)
 *                     if not m:             # <<<<<<<<<<<<<<
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 * 
 */
              }

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":825
 *                         raise BitmapFontException("Bitmap font descriptor file invalid format")
 * 
 *                     first = int(m.group("first"))             # <<<<<<<<<<<<<<
 *                     second = int(m.group("second"))
 * 
 */
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 825, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_3 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
                __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
                if (likely(__pyx_t_3)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                  __Pyx_INCREF(__pyx_t_3);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_1, function);
                }
              }
              __pyx_t_15 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_u_first) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_u_first);
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 825, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 825, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 825, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_v_first = __pyx_t_10;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":826
 * 
 *                     first = int(m.group("first"))
 *                     second = int(m.group("second"))             # <<<<<<<<<<<<<<
 * 
 *                     if first not in self.kernings:
 */
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 826, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_3 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_15))) {
                __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_15);
                if (likely(__pyx_t_3)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                  __Pyx_INCREF(__pyx_t_3);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_15, function);
                }
              }
              __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_3, __pyx_n_u_second) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_u_second);
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 826, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_15 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 826, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_15); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 826, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_v_second = __pyx_t_10;

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":828
 *                     second = int(m.group("second"))
 * 
 *                     if first not in self.kernings:             # <<<<<<<<<<<<<<
 *                         self.kernings[first] = {}
 * 
 */
              __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_first); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 828, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              if (unlikely(__pyx_v_self->kernings == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                __PYX_ERR(1, 828, __pyx_L14_error)
              }
              __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_t_15, __pyx_v_self->kernings, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 828, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_5 = (__pyx_t_4 != 0);
              if (__pyx_t_5) {

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":829
 * 
 *                     if first not in self.kernings:
 *                         self.kernings[first] = {}             # <<<<<<<<<<<<<<
 * 
 *                     if second not in self.kernings[first]:
 */
                __pyx_t_15 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 829, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_15);
                if (unlikely(__pyx_v_self->kernings == Py_None)) {
                  PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                  __PYX_ERR(1, 829, __pyx_L14_error)
                }
                __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_first); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 829, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_1);
                if (unlikely(PyDict_SetItem(__pyx_v_self->kernings, __pyx_t_1, __pyx_t_15) < 0)) __PYX_ERR(1, 829, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":828
 *                     second = int(m.group("second"))
 * 
 *                     if first not in self.kernings:             # <<<<<<<<<<<<<<
 *                         self.kernings[first] = {}
 * 
 */
              }

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":831
 *                         self.kernings[first] = {}
 * 
 *                     if second not in self.kernings[first]:             # <<<<<<<<<<<<<<
 *                         self.kernings[first][second] = int(m.group("amount"))
 * 
 */
              __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_second); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 831, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_15);
              if (unlikely(__pyx_v_self->kernings == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(1, 831, __pyx_L14_error)
              }
              __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_first); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 831, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_self->kernings, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 831, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_t_15, __pyx_t_3, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 831, __pyx_L14_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_4 = (__pyx_t_5 != 0);
              if (__pyx_t_4) {

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":832
 * 
 *                     if second not in self.kernings[first]:
 *                         self.kernings[first][second] = int(m.group("amount"))             # <<<<<<<<<<<<<<
 * 
 *     def _load_descriptor_xml(self, xml_tree: ElementTree):
 */
                __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_m, __pyx_n_s_group); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 832, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_1 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_15))) {
                  __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_15);
                  if (likely(__pyx_t_1)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                    __Pyx_INCREF(__pyx_t_1);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_15, function);
                  }
                }
                __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_1, __pyx_n_u_amount) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_n_u_amount);
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 832, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __pyx_t_15 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 832, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_15);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                if (unlikely(__pyx_v_self->kernings == Py_None)) {
                  PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                  __PYX_ERR(1, 832, __pyx_L14_error)
                }
                __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_first); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 832, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->kernings, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 832, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                if (unlikely(__Pyx_SetItemInt(__pyx_t_1, __pyx_v_second, __pyx_t_15, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(1, 832, __pyx_L14_error)
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

                /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":831
 *                         self.kernings[first] = {}
 * 
 *                     if second not in self.kernings[first]:             # <<<<<<<<<<<<<<
 *                         self.kernings[first][second] = int(m.group("amount"))
 * 
 */
              }

              /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":819
 *                     pass
 * 
 *                 elif line.startswith("kerning"):             # <<<<<<<<<<<<<<
 *                     m = re.search(r"kerning\s+first=(?P<first>[0-9]{1,3})\s+second=(?P<second>[0-9]{1,3})"
 *                                   r"\s+amount=(?P<amount>-?[0-9]{1,3})", line, flags=re.IGNORECASE)
 */
            }
            __pyx_L22:;

            /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":767
 * 
 *             # Loop over all the rows in the descriptor file
 *             for line in text_file:             # <<<<<<<<<<<<<<
 *                 if line.startswith("info"):
 *                     pass
 */
            __pyx_L20_continue:;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":764
 * 
 *         # Open the descriptor file
 *         with open(descriptor_file) as text_file:             # <<<<<<<<<<<<<<
 * 
 *             # Loop over all the rows in the descriptor file
 */
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L19_try_end;
        __pyx_L14_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont._load_descriptor_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_15, &__pyx_t_1) < 0) __PYX_ERR(1, 764, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_15, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 764, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_3, NULL);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 764, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (__pyx_t_4 < 0) __PYX_ERR(1, 764, __pyx_L16_except_error)
          __pyx_t_5 = ((!(__pyx_t_4 != 0)) != 0);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_1);
            __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_15, __pyx_t_1);
            __pyx_t_2 = 0; __pyx_t_15 = 0; __pyx_t_1 = 0; 
            __PYX_ERR(1, 764, __pyx_L16_except_error)
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L15_exception_handled;
        }
        __pyx_L16_except_error:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L15_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_11);
        __pyx_L19_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_9) {
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_tuple__4, NULL);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 764, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        goto __pyx_L13;
      }
      __pyx_L13:;
    }
    goto __pyx_L34;
    __pyx_L10_error:;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L1_error;
    __pyx_L34:;
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":739
 *             y += row_height
 * 
 *     def _load_descriptor_file(self, str descriptor_file):             # <<<<<<<<<<<<<<
 *         """Load the descriptor from the specified file.  Both XML and text
 *         formats are supported (not binary currently).
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont._load_descriptor_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_xml_tree);
  __Pyx_XDECREF(__pyx_v_text_file);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XDECREF((PyObject *)__pyx_v_character);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":834
 *                         self.kernings[first][second] = int(m.group("amount"))
 * 
 *     def _load_descriptor_xml(self, xml_tree: ElementTree):             # <<<<<<<<<<<<<<
 *         """Loads the descriptor information from the XML tree."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_17_load_descriptor_xml(PyObject *__pyx_v_self, PyObject *__pyx_v_xml_tree); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_16_load_descriptor_xml[] = "BitmapFont._load_descriptor_xml(self, xml_tree: ElementTree)\nLoads the descriptor information from the XML tree.";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_17_load_descriptor_xml(PyObject *__pyx_v_self, PyObject *__pyx_v_xml_tree) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_load_descriptor_xml (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_16_load_descriptor_xml(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_xml_tree));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_16_load_descriptor_xml(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_xml_tree) {
  int __pyx_v_first;
  int __pyx_v_second;
  PyObject *__pyx_v_root = NULL;
  PyObject *__pyx_v_info = NULL;
  PyObject *__pyx_v_common = NULL;
  PyObject *__pyx_v_chars = NULL;
  PyObject *__pyx_v_text_char = NULL;
  struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_character = NULL;
  PyObject *__pyx_v_kernings = NULL;
  PyObject *__pyx_v_kerning = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_load_descriptor_xml", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":837
 *         """Loads the descriptor information from the XML tree."""
 * 
 *         cdef int first = 0             # <<<<<<<<<<<<<<
 *         cdef int second = 0
 *         root = xml_tree.getroot()
 */
  __pyx_v_first = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":838
 * 
 *         cdef int first = 0
 *         cdef int second = 0             # <<<<<<<<<<<<<<
 *         root = xml_tree.getroot()
 *         if root is None:
 */
  __pyx_v_second = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":839
 *         cdef int first = 0
 *         cdef int second = 0
 *         root = xml_tree.getroot()             # <<<<<<<<<<<<<<
 *         if root is None:
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_xml_tree, __pyx_n_s_getroot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_root = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":840
 *         cdef int second = 0
 *         root = xml_tree.getroot()
 *         if root is None:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 */
  __pyx_t_4 = (__pyx_v_root == Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (unlikely(__pyx_t_5)) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":841
 *         root = xml_tree.getroot()
 *         if root is None:
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")             # <<<<<<<<<<<<<<
 * 
 *         # info
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u_Bitmap_font_descriptor_file_inva_2) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_Bitmap_font_descriptor_file_inva_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 841, __pyx_L1_error)

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":840
 *         cdef int second = 0
 *         root = xml_tree.getroot()
 *         if root is None:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":844
 * 
 *         # info
 *         info = root.find('info')             # <<<<<<<<<<<<<<
 *         if info is None:
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_root, __pyx_n_s_find); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_n_u_info) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_info);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":845
 *         # info
 *         info = root.find('info')
 *         if info is None:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 */
  __pyx_t_5 = (__pyx_v_info == Py_None);
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (unlikely(__pyx_t_4)) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":846
 *         info = root.find('info')
 *         if info is None:
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")             # <<<<<<<<<<<<<<
 * 
 *         # common
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 846, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u_Bitmap_font_descriptor_file_inva_2) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_Bitmap_font_descriptor_file_inva_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 846, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 846, __pyx_L1_error)

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":845
 *         # info
 *         info = root.find('info')
 *         if info is None:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":849
 * 
 *         # common
 *         common = root.find('common')             # <<<<<<<<<<<<<<
 *         if common is None:
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_root, __pyx_n_s_find); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_n_u_common) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_common);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_common = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":850
 *         # common
 *         common = root.find('common')
 *         if common is None:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 */
  __pyx_t_4 = (__pyx_v_common == Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (unlikely(__pyx_t_5)) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":851
 *         common = root.find('common')
 *         if common is None:
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")             # <<<<<<<<<<<<<<
 * 
 *         self.line_height = int(common.attrib["lineHeight"])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u_Bitmap_font_descriptor_file_inva_2) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_Bitmap_font_descriptor_file_inva_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 851, __pyx_L1_error)

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":850
 *         # common
 *         common = root.find('common')
 *         if common is None:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":853
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 *         self.line_height = int(common.attrib["lineHeight"])             # <<<<<<<<<<<<<<
 *         self.base = int(common.attrib["base"])
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_common, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_lineHeight); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->line_height = __pyx_t_6;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":854
 * 
 *         self.line_height = int(common.attrib["lineHeight"])
 *         self.base = int(common.attrib["base"])             # <<<<<<<<<<<<<<
 * 
 *         # characters
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_common, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 854, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->base = __pyx_t_6;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":857
 * 
 *         # characters
 *         chars = root.find('chars')             # <<<<<<<<<<<<<<
 *         if chars is None:
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_root, __pyx_n_s_find); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_n_u_chars) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_chars);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_chars = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":858
 *         # characters
 *         chars = root.find('chars')
 *         if chars is None:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 */
  __pyx_t_5 = (__pyx_v_chars == Py_None);
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (unlikely(__pyx_t_4)) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":859
 *         chars = root.find('chars')
 *         if chars is None:
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")             # <<<<<<<<<<<<<<
 * 
 *         for text_char in chars.findall('char'):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BitmapFontException); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u_Bitmap_font_descriptor_file_inva_2) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_Bitmap_font_descriptor_file_inva_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 859, __pyx_L1_error)

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":858
 *         # characters
 *         chars = root.find('chars')
 *         if chars is None:             # <<<<<<<<<<<<<<
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":861
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 *         for text_char in chars.findall('char'):             # <<<<<<<<<<<<<<
 *             character = BitmapFontCharacter()
 *             character.id = int(text_char.attrib["id"])
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_chars, __pyx_n_s_findall); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_n_u_char) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_char);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 861, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 861, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 861, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 861, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 861, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_8(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 861, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_text_char, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":862
 * 
 *         for text_char in chars.findall('char'):
 *             character = BitmapFontCharacter()             # <<<<<<<<<<<<<<
 *             character.id = int(text_char.attrib["id"])
 *             character.rect.x = int(text_char.attrib["x"])
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_character, ((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":863
 *         for text_char in chars.findall('char'):
 *             character = BitmapFontCharacter()
 *             character.id = int(text_char.attrib["id"])             # <<<<<<<<<<<<<<
 *             character.rect.x = int(text_char.attrib["x"])
 *             character.rect.y = int(text_char.attrib["y"])
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_text_char, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 863, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_character->id = __pyx_t_6;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":864
 *             character = BitmapFontCharacter()
 *             character.id = int(text_char.attrib["id"])
 *             character.rect.x = int(text_char.attrib["x"])             # <<<<<<<<<<<<<<
 *             character.rect.y = int(text_char.attrib["y"])
 *             character.rect.w = int(text_char.attrib["width"])
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_text_char, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 864, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_character->rect.x = __pyx_t_6;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":865
 *             character.id = int(text_char.attrib["id"])
 *             character.rect.x = int(text_char.attrib["x"])
 *             character.rect.y = int(text_char.attrib["y"])             # <<<<<<<<<<<<<<
 *             character.rect.w = int(text_char.attrib["width"])
 *             character.rect.h = int(text_char.attrib["height"])
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_text_char, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 865, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_character->rect.y = __pyx_t_6;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":866
 *             character.rect.x = int(text_char.attrib["x"])
 *             character.rect.y = int(text_char.attrib["y"])
 *             character.rect.w = int(text_char.attrib["width"])             # <<<<<<<<<<<<<<
 *             character.rect.h = int(text_char.attrib["height"])
 *             character.xadvance = int(text_char.attrib["xadvance"])
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_text_char, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_width); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 866, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_character->rect.w = __pyx_t_6;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":867
 *             character.rect.y = int(text_char.attrib["y"])
 *             character.rect.w = int(text_char.attrib["width"])
 *             character.rect.h = int(text_char.attrib["height"])             # <<<<<<<<<<<<<<
 *             character.xadvance = int(text_char.attrib["xadvance"])
 *             character.xoffset = int(text_char.attrib["xoffset"])
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_text_char, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 867, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_height); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 867, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 867, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 867, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_character->rect.h = __pyx_t_6;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":868
 *             character.rect.w = int(text_char.attrib["width"])
 *             character.rect.h = int(text_char.attrib["height"])
 *             character.xadvance = int(text_char.attrib["xadvance"])             # <<<<<<<<<<<<<<
 *             character.xoffset = int(text_char.attrib["xoffset"])
 *             character.yoffset = int(text_char.attrib["yoffset"])
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_text_char, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_xadvance); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 868, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_character->xadvance = __pyx_t_6;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":869
 *             character.rect.h = int(text_char.attrib["height"])
 *             character.xadvance = int(text_char.attrib["xadvance"])
 *             character.xoffset = int(text_char.attrib["xoffset"])             # <<<<<<<<<<<<<<
 *             character.yoffset = int(text_char.attrib["yoffset"])
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_text_char, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_xoffset); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 869, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_character->xoffset = __pyx_t_6;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":870
 *             character.xadvance = int(text_char.attrib["xadvance"])
 *             character.xoffset = int(text_char.attrib["xoffset"])
 *             character.yoffset = int(text_char.attrib["yoffset"])             # <<<<<<<<<<<<<<
 * 
 *             self.characters[character.id] = character
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_text_char, __pyx_n_s_attrib); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 870, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_yoffset); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 870, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 870, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 870, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_character->yoffset = __pyx_t_6;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":872
 *             character.yoffset = int(text_char.attrib["yoffset"])
 * 
 *             self.characters[character.id] = character             # <<<<<<<<<<<<<<
 * 
 *         # kerning
 */
    if (unlikely(__pyx_v_self->characters == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 872, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_character->id); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyDict_SetItem(__pyx_v_self->characters, __pyx_t_1, ((PyObject *)__pyx_v_character)) < 0)) __PYX_ERR(1, 872, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":861
 *             raise BitmapFontException("Bitmap font descriptor file invalid XML format")
 * 
 *         for text_char in chars.findall('char'):             # <<<<<<<<<<<<<<
 *             character = BitmapFontCharacter()
 *             character.id = int(text_char.attrib["id"])
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":875
 * 
 *         # kerning
 *         kernings = root.find('kernings')             # <<<<<<<<<<<<<<
 *         if kernings is not None:
 *             for kerning in kernings.findall('kerning'):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_root, __pyx_n_s_find); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_u_kernings) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_u_kernings);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_kernings = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":876
 *         # kerning
 *         kernings = root.find('kernings')
 *         if kernings is not None:             # <<<<<<<<<<<<<<
 *             for kerning in kernings.findall('kerning'):
 *                 first = int(kerning.attrib["first"])
 */
  __pyx_t_4 = (__pyx_v_kernings != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":877
 *         kernings = root.find('kernings')
 *         if kernings is not None:
 *             for kerning in kernings.findall('kerning'):             # <<<<<<<<<<<<<<
 *                 first = int(kerning.attrib["first"])
 *                 second = int(kerning.attrib["second"])
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_kernings, __pyx_n_s_findall); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 877, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_n_u_kerning) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_u_kerning);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 877, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 877, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 877, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 877, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 877, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 877, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_8(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(1, 877, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_kerning, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":878
 *         if kernings is not None:
 *             for kerning in kernings.findall('kerning'):
 *                 first = int(kerning.attrib["first"])             # <<<<<<<<<<<<<<
 *                 second = int(kerning.attrib["second"])
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_kerning, __pyx_n_s_attrib); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_first); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 878, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_first = __pyx_t_6;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":879
 *             for kerning in kernings.findall('kerning'):
 *                 first = int(kerning.attrib["first"])
 *                 second = int(kerning.attrib["second"])             # <<<<<<<<<<<<<<
 * 
 *                 if first not in self.kernings:
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_kerning, __pyx_n_s_attrib); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_second); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 879, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_second = __pyx_t_6;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":881
 *                 second = int(kerning.attrib["second"])
 * 
 *                 if first not in self.kernings:             # <<<<<<<<<<<<<<
 *                     self.kernings[first] = {}
 * 
 */
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_first); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__pyx_v_self->kernings == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(1, 881, __pyx_L1_error)
      }
      __pyx_t_5 = (__Pyx_PyDict_ContainsTF(__pyx_t_2, __pyx_v_self->kernings, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 881, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_4 = (__pyx_t_5 != 0);
      if (__pyx_t_4) {

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":882
 * 
 *                 if first not in self.kernings:
 *                     self.kernings[first] = {}             # <<<<<<<<<<<<<<
 * 
 *                 self.kernings[first][second] = int(kerning.attrib["amount"])
 */
        __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 882, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(__pyx_v_self->kernings == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(1, 882, __pyx_L1_error)
        }
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_first); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 882, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(PyDict_SetItem(__pyx_v_self->kernings, __pyx_t_3, __pyx_t_2) < 0)) __PYX_ERR(1, 882, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":881
 *                 second = int(kerning.attrib["second"])
 * 
 *                 if first not in self.kernings:             # <<<<<<<<<<<<<<
 *                     self.kernings[first] = {}
 * 
 */
      }

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":884
 *                     self.kernings[first] = {}
 * 
 *                 self.kernings[first][second] = int(kerning.attrib["amount"])             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_kerning, __pyx_n_s_attrib); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__pyx_v_self->kernings == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(1, 884, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_first); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = __Pyx_PyDict_GetItem(__pyx_v_self->kernings, __pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_SetItemInt(__pyx_t_9, __pyx_v_second, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(1, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":877
 *         kernings = root.find('kernings')
 *         if kernings is not None:
 *             for kerning in kernings.findall('kerning'):             # <<<<<<<<<<<<<<
 *                 first = int(kerning.attrib["first"])
 *                 second = int(kerning.attrib["second"])
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":876
 *         # kerning
 *         kernings = root.find('kernings')
 *         if kernings is not None:             # <<<<<<<<<<<<<<
 *             for kerning in kernings.findall('kerning'):
 *                 first = int(kerning.attrib["first"])
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":834
 *                         self.kernings[first][second] = int(m.group("amount"))
 * 
 *     def _load_descriptor_xml(self, xml_tree: ElementTree):             # <<<<<<<<<<<<<<
 *         """Loads the descriptor information from the XML tree."""
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont._load_descriptor_xml", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_root);
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_common);
  __Pyx_XDECREF(__pyx_v_chars);
  __Pyx_XDECREF(__pyx_v_text_char);
  __Pyx_XDECREF((PyObject *)__pyx_v_character);
  __Pyx_XDECREF(__pyx_v_kernings);
  __Pyx_XDECREF(__pyx_v_kerning);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":887
 * 
 * 
 *     def get_descent(self):             # <<<<<<<<<<<<<<
 *         return self.base - self.line_height
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_19get_descent(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_18get_descent[] = "BitmapFont.get_descent(self)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_19get_descent(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_descent (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_18get_descent(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_18get_descent(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_descent", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":888
 * 
 *     def get_descent(self):
 *         return self.base - self.line_height             # <<<<<<<<<<<<<<
 * 
 *     def get_ascent(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int((__pyx_v_self->base - __pyx_v_self->line_height)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":887
 * 
 * 
 *     def get_descent(self):             # <<<<<<<<<<<<<<
 *         return self.base - self.line_height
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.get_descent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":890
 *         return self.base - self.line_height
 * 
 *     def get_ascent(self):             # <<<<<<<<<<<<<<
 *         return self.base
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_21get_ascent(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_20get_ascent[] = "BitmapFont.get_ascent(self)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_21get_ascent(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_ascent (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_20get_ascent(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_20get_ascent(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_ascent", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":891
 * 
 *     def get_ascent(self):
 *         return self.base             # <<<<<<<<<<<<<<
 * 
 *     def get_extents(self, str text, font_kerning=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":890
 *         return self.base - self.line_height
 * 
 *     def get_ascent(self):             # <<<<<<<<<<<<<<
 *         return self.base
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.get_ascent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":893
 *         return self.base
 * 
 *     def get_extents(self, str text, font_kerning=True):             # <<<<<<<<<<<<<<
 *         """Calculate the actual size of the specified text when rendered."""
 *         cdef int left_x = 0
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_23get_extents(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_22get_extents[] = "BitmapFont.get_extents(self, unicode text, font_kerning=True)\nCalculate the actual size of the specified text when rendered.";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_23get_extents(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_text = 0;
  PyObject *__pyx_v_font_kerning = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_extents (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_text,&__pyx_n_s_font_kerning,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_text)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_font_kerning);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_extents") < 0)) __PYX_ERR(1, 893, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_text = ((PyObject*)values[0]);
    __pyx_v_font_kerning = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_extents", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 893, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.get_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_text), (&PyUnicode_Type), 1, "text", 1))) __PYX_ERR(1, 893, __pyx_L1_error)
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_22get_extents(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), __pyx_v_text, __pyx_v_font_kerning);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_22get_extents(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_font_kerning) {
  int __pyx_v_left_x;
  int __pyx_v_right_x;
  int __pyx_v_cursor_x;
  int __pyx_v_previous_char;
  int __pyx_v_current_char;
  int __pyx_v_use_kerning;
  struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *__pyx_v_char_info = 0;
  Py_ssize_t __pyx_v_index;
  PyObject *__pyx_v_text_char = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  long __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_extents", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":895
 *     def get_extents(self, str text, font_kerning=True):
 *         """Calculate the actual size of the specified text when rendered."""
 *         cdef int left_x = 0             # <<<<<<<<<<<<<<
 *         cdef int right_x = 0
 *         cdef int cursor_x = 0
 */
  __pyx_v_left_x = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":896
 *         """Calculate the actual size of the specified text when rendered."""
 *         cdef int left_x = 0
 *         cdef int right_x = 0             # <<<<<<<<<<<<<<
 *         cdef int cursor_x = 0
 *         cdef int previous_char = -1
 */
  __pyx_v_right_x = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":897
 *         cdef int left_x = 0
 *         cdef int right_x = 0
 *         cdef int cursor_x = 0             # <<<<<<<<<<<<<<
 *         cdef int previous_char = -1
 *         cdef int current_char
 */
  __pyx_v_cursor_x = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":898
 *         cdef int right_x = 0
 *         cdef int cursor_x = 0
 *         cdef int previous_char = -1             # <<<<<<<<<<<<<<
 *         cdef int current_char
 *         cdef bint use_kerning = font_kerning
 */
  __pyx_v_previous_char = -1;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":900
 *         cdef int previous_char = -1
 *         cdef int current_char
 *         cdef bint use_kerning = font_kerning             # <<<<<<<<<<<<<<
 *         cdef BitmapFontCharacter char_info
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_font_kerning); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 900, __pyx_L1_error)
  __pyx_v_use_kerning = __pyx_t_1;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":904
 * 
 *         # loop over characters getting character information
 *         for index, text_char in enumerate(text):             # <<<<<<<<<<<<<<
 *             current_char = ord(text_char)
 *             if current_char in self.characters:
 */
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v_text == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(1, 904, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_text);
  __pyx_t_3 = __pyx_v_text;
  __pyx_t_8 = __Pyx_init_unicode_iteration(__pyx_t_3, (&__pyx_t_5), (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(1, 904, __pyx_L1_error)
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_5; __pyx_t_9++) {
    __pyx_t_4 = __pyx_t_9;
    __pyx_t_10 = PyUnicode_FromOrdinal(__Pyx_PyUnicode_READ(__pyx_t_7, __pyx_t_6, __pyx_t_4)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_XDECREF_SET(__pyx_v_text_char, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":905
 *         # loop over characters getting character information
 *         for index, text_char in enumerate(text):
 *             current_char = ord(text_char)             # <<<<<<<<<<<<<<
 *             if current_char in self.characters:
 *                 char_info = self.characters[current_char]
 */
    __pyx_v_index = __pyx_t_2;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":904
 * 
 *         # loop over characters getting character information
 *         for index, text_char in enumerate(text):             # <<<<<<<<<<<<<<
 *             current_char = ord(text_char)
 *             if current_char in self.characters:
 */
    __pyx_t_2 = (__pyx_t_2 + 1);

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":905
 *         # loop over characters getting character information
 *         for index, text_char in enumerate(text):
 *             current_char = ord(text_char)             # <<<<<<<<<<<<<<
 *             if current_char in self.characters:
 *                 char_info = self.characters[current_char]
 */
    __pyx_t_11 = __Pyx_PyObject_Ord(__pyx_v_text_char); if (unlikely(__pyx_t_11 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(1, 905, __pyx_L1_error)
    __pyx_v_current_char = __pyx_t_11;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":906
 *         for index, text_char in enumerate(text):
 *             current_char = ord(text_char)
 *             if current_char in self.characters:             # <<<<<<<<<<<<<<
 *                 char_info = self.characters[current_char]
 * 
 */
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_current_char); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (unlikely(__pyx_v_self->characters == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(1, 906, __pyx_L1_error)
    }
    __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_t_10, __pyx_v_self->characters, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 906, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_12 = (__pyx_t_1 != 0);
    if (__pyx_t_12) {

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":907
 *             current_char = ord(text_char)
 *             if current_char in self.characters:
 *                 char_info = self.characters[current_char]             # <<<<<<<<<<<<<<
 * 
 *                 # capture the left coordinate of the first character
 */
      if (unlikely(__pyx_v_self->characters == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(1, 907, __pyx_L1_error)
      }
      __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_current_char); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_13 = __Pyx_PyDict_GetItem(__pyx_v_self->characters, __pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (!(likely(((__pyx_t_13) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_13, __pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter))))) __PYX_ERR(1, 907, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_char_info, ((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter *)__pyx_t_13));
      __pyx_t_13 = 0;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":910
 * 
 *                 # capture the left coordinate of the first character
 *                 if index == 0:             # <<<<<<<<<<<<<<
 *                     left_x = char_info.xoffset
 * 
 */
      __pyx_t_12 = ((__pyx_v_index == 0) != 0);
      if (__pyx_t_12) {

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":911
 *                 # capture the left coordinate of the first character
 *                 if index == 0:
 *                     left_x = char_info.xoffset             # <<<<<<<<<<<<<<
 * 
 *                 # apply kerning (if applicable)
 */
        __pyx_t_8 = __pyx_v_char_info->xoffset;
        __pyx_v_left_x = __pyx_t_8;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":910
 * 
 *                 # capture the left coordinate of the first character
 *                 if index == 0:             # <<<<<<<<<<<<<<
 *                     left_x = char_info.xoffset
 * 
 */
      }

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":914
 * 
 *                 # apply kerning (if applicable)
 *                 if use_kerning and previous_char in self.kernings and current_char in self.kernings[previous_char]:             # <<<<<<<<<<<<<<
 *                     cursor_x += self.kernings[previous_char][current_char]
 * 
 */
      __pyx_t_1 = (__pyx_v_use_kerning != 0);
      if (__pyx_t_1) {
      } else {
        __pyx_t_12 = __pyx_t_1;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_previous_char); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (unlikely(__pyx_v_self->kernings == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(1, 914, __pyx_L1_error)
      }
      __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_t_13, __pyx_v_self->kernings, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_14 = (__pyx_t_1 != 0);
      if (__pyx_t_14) {
      } else {
        __pyx_t_12 = __pyx_t_14;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_current_char); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (unlikely(__pyx_v_self->kernings == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(1, 914, __pyx_L1_error)
      }
      __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_previous_char); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_15 = __Pyx_PyDict_GetItem(__pyx_v_self->kernings, __pyx_t_10); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_14 = (__Pyx_PySequence_ContainsTF(__pyx_t_13, __pyx_t_15, Py_EQ)); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_1 = (__pyx_t_14 != 0);
      __pyx_t_12 = __pyx_t_1;
      __pyx_L8_bool_binop_done:;
      if (__pyx_t_12) {

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":915
 *                 # apply kerning (if applicable)
 *                 if use_kerning and previous_char in self.kernings and current_char in self.kernings[previous_char]:
 *                     cursor_x += self.kernings[previous_char][current_char]             # <<<<<<<<<<<<<<
 * 
 *                 # determine right coordinate of current character
 */
        __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_cursor_x); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (unlikely(__pyx_v_self->kernings == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(1, 915, __pyx_L1_error)
        }
        __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_previous_char); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_self->kernings, __pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_10, __pyx_v_current_char, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = PyNumber_InPlaceAdd(__pyx_t_15, __pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_10); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 915, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_v_cursor_x = __pyx_t_8;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":914
 * 
 *                 # apply kerning (if applicable)
 *                 if use_kerning and previous_char in self.kernings and current_char in self.kernings[previous_char]:             # <<<<<<<<<<<<<<
 *                     cursor_x += self.kernings[previous_char][current_char]
 * 
 */
      }

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":918
 * 
 *                 # determine right coordinate of current character
 *                 right_x = cursor_x + char_info.xoffset + char_info.rect.w             # <<<<<<<<<<<<<<
 * 
 *                 # advance the cursor to measure the next character
 */
      __pyx_v_right_x = ((__pyx_v_cursor_x + __pyx_v_char_info->xoffset) + __pyx_v_char_info->rect.w);

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":921
 * 
 *                 # advance the cursor to measure the next character
 *                 cursor_x += char_info.xadvance             # <<<<<<<<<<<<<<
 * 
 *             previous_char = current_char
 */
      __pyx_v_cursor_x = (__pyx_v_cursor_x + __pyx_v_char_info->xadvance);

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":906
 *         for index, text_char in enumerate(text):
 *             current_char = ord(text_char)
 *             if current_char in self.characters:             # <<<<<<<<<<<<<<
 *                 char_info = self.characters[current_char]
 * 
 */
    }

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":923
 *                 cursor_x += char_info.xadvance
 * 
 *             previous_char = current_char             # <<<<<<<<<<<<<<
 * 
 *         return right_x - left_x, self.line_height
 */
    __pyx_v_previous_char = __pyx_v_current_char;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":925
 *             previous_char = current_char
 * 
 *         return right_x - left_x, self.line_height             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_10 = __Pyx_PyInt_From_int((__pyx_v_right_x - __pyx_v_left_x)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_self->line_height); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_13);
  __pyx_t_10 = 0;
  __pyx_t_13 = 0;
  __pyx_r = __pyx_t_15;
  __pyx_t_15 = 0;
  goto __pyx_L0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":893
 *         return self.base
 * 
 *     def get_extents(self, str text, font_kerning=True):             # <<<<<<<<<<<<<<
 *         """Calculate the actual size of the specified text when rendered."""
 *         cdef int left_x = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.get_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_char_info);
  __Pyx_XDECREF(__pyx_v_text_char);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":633
 * 
 * cdef class BitmapFont:
 *     cdef public str face             # <<<<<<<<<<<<<<
 *     cdef public bint bold
 *     cdef public bint italic
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->face);
  __pyx_r = __pyx_v_self->face;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 633, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->face);
  __Pyx_DECREF(__pyx_v_self->face);
  __pyx_v_self->face = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.face.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_4__del__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_4__del__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->face);
  __Pyx_DECREF(__pyx_v_self->face);
  __pyx_v_self->face = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":634
 * cdef class BitmapFont:
 *     cdef public str face
 *     cdef public bint bold             # <<<<<<<<<<<<<<
 *     cdef public bint italic
 *     cdef public int padding
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->bold); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.bold.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 634, __pyx_L1_error)
  __pyx_v_self->bold = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.bold.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":635
 *     cdef public str face
 *     cdef public bint bold
 *     cdef public bint italic             # <<<<<<<<<<<<<<
 *     cdef public int padding
 *     cdef public int spacing
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->italic); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.italic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 635, __pyx_L1_error)
  __pyx_v_self->italic = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.italic.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":636
 *     cdef public bint bold
 *     cdef public bint italic
 *     cdef public int padding             # <<<<<<<<<<<<<<
 *     cdef public int spacing
 *     cdef public int outline
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->padding); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.padding.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 636, __pyx_L1_error)
  __pyx_v_self->padding = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.padding.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":637
 *     cdef public bint italic
 *     cdef public int padding
 *     cdef public int spacing             # <<<<<<<<<<<<<<
 *     cdef public int outline
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->spacing); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.spacing.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 637, __pyx_L1_error)
  __pyx_v_self->spacing = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.spacing.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":638
 *     cdef public int padding
 *     cdef public int spacing
 *     cdef public int outline             # <<<<<<<<<<<<<<
 * 
 *     cdef public int line_height
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->outline); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.outline.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 638, __pyx_L1_error)
  __pyx_v_self->outline = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.outline.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":640
 *     cdef public int outline
 * 
 *     cdef public int line_height             # <<<<<<<<<<<<<<
 *     cdef public int base
 *     cdef public int scale_w
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->line_height); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.line_height.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 640, __pyx_L1_error)
  __pyx_v_self->line_height = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.line_height.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":641
 * 
 *     cdef public int line_height
 *     cdef public int base             # <<<<<<<<<<<<<<
 *     cdef public int scale_w
 *     cdef public int scale_h
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.base.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 641, __pyx_L1_error)
  __pyx_v_self->base = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.base.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":642
 *     cdef public int line_height
 *     cdef public int base
 *     cdef public int scale_w             # <<<<<<<<<<<<<<
 *     cdef public int scale_h
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->scale_w); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.scale_w.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 642, __pyx_L1_error)
  __pyx_v_self->scale_w = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.scale_w.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":643
 *     cdef public int base
 *     cdef public int scale_w
 *     cdef public int scale_h             # <<<<<<<<<<<<<<
 * 
 *     cdef dict characters
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h___get__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h___get__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->scale_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.scale_h.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h_2__set__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h_2__set__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 643, __pyx_L1_error)
  __pyx_v_self->scale_h = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.scale_h.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_25__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_24__reduce_cython__[] = "BitmapFont.__reduce_cython__(self)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_25__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_24__reduce_cython__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_24__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_27__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_26__setstate_cython__[] = "BitmapFont.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_27__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_26__setstate_cython__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_26__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font.BitmapFont.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":932
 *     cdef int w, h
 * 
 *     def __cinit__(self, w, h):             # <<<<<<<<<<<<<<
 *         self.surface = NULL
 *         self.w = w
 */

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_w = 0;
  PyObject *__pyx_v_h = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_w,&__pyx_n_s_h,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(1, 932, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 932, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_w = values[0];
    __pyx_v_h = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 932, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer___cinit__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *)__pyx_v_self), __pyx_v_w, __pyx_v_h);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer___cinit__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self, PyObject *__pyx_v_w, PyObject *__pyx_v_h) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":933
 * 
 *     def __cinit__(self, w, h):
 *         self.surface = NULL             # <<<<<<<<<<<<<<
 *         self.w = w
 *         self.h = h
 */
  __pyx_v_self->surface = NULL;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":934
 *     def __cinit__(self, w, h):
 *         self.surface = NULL
 *         self.w = w             # <<<<<<<<<<<<<<
 *         self.h = h
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_w); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 934, __pyx_L1_error)
  __pyx_v_self->w = __pyx_t_1;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":935
 *         self.surface = NULL
 *         self.w = w
 *         self.h = h             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, w, h):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_h); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 935, __pyx_L1_error)
  __pyx_v_self->h = __pyx_t_1;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":932
 *     cdef int w, h
 * 
 *     def __cinit__(self, w, h):             # <<<<<<<<<<<<<<
 *         self.surface = NULL
 *         self.w = w
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":937
 *         self.h = h
 * 
 *     def __init__(self, w, h):             # <<<<<<<<<<<<<<
 *         # XXX check on OSX to see if little endian/big endian make a difference here.
 *         self.surface = SDL_CreateRGBSurface(0, w, h, 32, 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000)
 */

/* Python wrapper */
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_w = 0;
  PyObject *__pyx_v_h = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_w,&__pyx_n_s_h,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(1, 937, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 937, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_w = values[0];
    __pyx_v_h = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 937, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_2__init__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *)__pyx_v_self), __pyx_v_w, __pyx_v_h);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_2__init__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self, PyObject *__pyx_v_w, PyObject *__pyx_v_h) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":939
 *     def __init__(self, w, h):
 *         # XXX check on OSX to see if little endian/big endian make a difference here.
 *         self.surface = SDL_CreateRGBSurface(0, w, h, 32, 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000)             # <<<<<<<<<<<<<<
 *         memset(self.surface.pixels, 0, w * h * 4)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_w); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 939, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_h); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 939, __pyx_L1_error)
  __pyx_v_self->surface = SDL_CreateRGBSurface(0, __pyx_t_1, __pyx_t_2, 32, 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":940
 *         # XXX check on OSX to see if little endian/big endian make a difference here.
 *         self.surface = SDL_CreateRGBSurface(0, w, h, 32, 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000)
 *         memset(self.surface.pixels, 0, w * h * 4)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_w, __pyx_v_h); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_3, __pyx_int_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 940, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  (void)(memset(__pyx_v_self->surface->pixels, 0, __pyx_t_5));

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":937
 *         self.h = h
 * 
 *     def __init__(self, w, h):             # <<<<<<<<<<<<<<
 *         # XXX check on OSX to see if little endian/big endian make a difference here.
 *         self.surface = SDL_CreateRGBSurface(0, w, h, 32, 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":942
 *         memset(self.surface.pixels, 0, w * h * 4)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.surface != NULL:
 *             SDL_FreeSurface(self.surface)
 */

/* Python wrapper */
static void __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_5__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_4__dealloc__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_4__dealloc__(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":943
 * 
 *     def __dealloc__(self):
 *         if self.surface != NULL:             # <<<<<<<<<<<<<<
 *             SDL_FreeSurface(self.surface)
 *             self.surface = NULL
 */
  __pyx_t_1 = ((__pyx_v_self->surface != NULL) != 0);
  if (__pyx_t_1) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":944
 *     def __dealloc__(self):
 *         if self.surface != NULL:
 *             SDL_FreeSurface(self.surface)             # <<<<<<<<<<<<<<
 *             self.surface = NULL
 * 
 */
    SDL_FreeSurface(__pyx_v_self->surface);

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":945
 *         if self.surface != NULL:
 *             SDL_FreeSurface(self.surface)
 *             self.surface = NULL             # <<<<<<<<<<<<<<
 * 
 *     def render(self, container, text, x, y):
 */
    __pyx_v_self->surface = NULL;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":943
 * 
 *     def __dealloc__(self):
 *         if self.surface != NULL:             # <<<<<<<<<<<<<<
 *             SDL_FreeSurface(self.surface)
 *             self.surface = NULL
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":942
 *         memset(self.surface.pixels, 0, w * h * 4)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.surface != NULL:
 *             SDL_FreeSurface(self.surface)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":947
 *             self.surface = NULL
 * 
 *     def render(self, container, text, x, y):             # <<<<<<<<<<<<<<
 *         """Render the text at the specified location."""
 *         cdef SDL_Rect cursor_rect
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_7render(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_6render[] = "_SurfaceContainer.render(self, container, text, x, y)\nRender the text at the specified location.";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_7render(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_container = 0;
  PyObject *__pyx_v_text = 0;
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_y = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("render (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_container,&__pyx_n_s_text,&__pyx_n_s_x,&__pyx_n_s_y,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_container)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_text)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("render", 1, 4, 4, 1); __PYX_ERR(1, 947, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("render", 1, 4, 4, 2); __PYX_ERR(1, 947, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("render", 1, 4, 4, 3); __PYX_ERR(1, 947, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "render") < 0)) __PYX_ERR(1, 947, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_container = values[0];
    __pyx_v_text = values[1];
    __pyx_v_x = values[2];
    __pyx_v_y = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("render", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 947, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer.render", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_6render(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *)__pyx_v_self), __pyx_v_container, __pyx_v_text, __pyx_v_x, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_6render(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self, PyObject *__pyx_v_container, PyObject *__pyx_v_text, PyObject *__pyx_v_x, PyObject *__pyx_v_y) {
  struct SDL_Rect __pyx_v_cursor_rect;
  struct SDL_Rect __pyx_v_source_rect;
  struct SDL_Rect __pyx_v_dest_rect;
  struct SDL_Surface *__pyx_v_source_image;
  PyObject *__pyx_v_characters = 0;
  PyObject *__pyx_v_kernings = 0;
  int __pyx_v_use_kerning;
  int __pyx_v_previous_char;
  int __pyx_v_current_char;
  PyObject *__pyx_v_asset = NULL;
  PyObject *__pyx_v_font = NULL;
  PyObject *__pyx_v_index = NULL;
  PyObject *__pyx_v_text_char = NULL;
  PyObject *__pyx_v_char_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  long __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  struct SDL_Rect __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("render", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":953
 *         cdef SDL_Rect dest_rect
 *         cdef SDL_Surface *source_image
 *         cdef dict characters = dict()             # <<<<<<<<<<<<<<
 *         cdef dict kernings = dict()
 *         cdef bint use_kerning = True
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_characters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":954
 *         cdef SDL_Surface *source_image
 *         cdef dict characters = dict()
 *         cdef dict kernings = dict()             # <<<<<<<<<<<<<<
 *         cdef bint use_kerning = True
 *         cdef int previous_char = -1
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_kernings = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":955
 *         cdef dict characters = dict()
 *         cdef dict kernings = dict()
 *         cdef bint use_kerning = True             # <<<<<<<<<<<<<<
 *         cdef int previous_char = -1
 *         cdef int current_char
 */
  __pyx_v_use_kerning = 1;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":956
 *         cdef dict kernings = dict()
 *         cdef bint use_kerning = True
 *         cdef int previous_char = -1             # <<<<<<<<<<<<<<
 *         cdef int current_char
 * 
 */
  __pyx_v_previous_char = -1;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":959
 *         cdef int current_char
 * 
 *         asset = container.get_font_asset()             # <<<<<<<<<<<<<<
 *         if asset is None:
 *             return
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_container, __pyx_n_s_get_font_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_asset = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":960
 * 
 *         asset = container.get_font_asset()
 *         if asset is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_4 = (__pyx_v_asset == Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":961
 *         asset = container.get_font_asset()
 *         if asset is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         font = asset.bitmap_font
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":960
 * 
 *         asset = container.get_font_asset()
 *         if asset is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":963
 *             return
 * 
 *         font = asset.bitmap_font             # <<<<<<<<<<<<<<
 *         if font is None:
 *             return
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_asset, __pyx_n_s_bitmap_font); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_font = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":964
 * 
 *         font = asset.bitmap_font
 *         if font is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_5 = (__pyx_v_font == Py_None);
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":965
 *         font = asset.bitmap_font
 *         if font is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         source_image = <SDL_Surface*>pycapsule.PyCapsule_GetPointer(font.get_image(), NULL)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":964
 * 
 *         font = asset.bitmap_font
 *         if font is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":967
 *             return
 * 
 *         source_image = <SDL_Surface*>pycapsule.PyCapsule_GetPointer(font.get_image(), NULL)             # <<<<<<<<<<<<<<
 *         if source_image == NULL:
 *             return
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_font, __pyx_n_s_get_image); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = PyCapsule_GetPointer(__pyx_t_1, NULL); if (unlikely(__pyx_t_6 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(1, 967, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_source_image = ((struct SDL_Surface *)__pyx_t_6);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":968
 * 
 *         source_image = <SDL_Surface*>pycapsule.PyCapsule_GetPointer(font.get_image(), NULL)
 *         if source_image == NULL:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_4 = ((__pyx_v_source_image == NULL) != 0);
  if (__pyx_t_4) {

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":969
 *         source_image = <SDL_Surface*>pycapsule.PyCapsule_GetPointer(font.get_image(), NULL)
 *         if source_image == NULL:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         """
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":968
 * 
 *         source_image = <SDL_Surface*>pycapsule.PyCapsule_GetPointer(font.get_image(), NULL)
 *         if source_image == NULL:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":992
 *         """
 * 
 *         cursor_rect.x = x             # <<<<<<<<<<<<<<
 *         cursor_rect.y = y
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_x); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 992, __pyx_L1_error)
  __pyx_v_cursor_rect.x = __pyx_t_7;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":993
 * 
 *         cursor_rect.x = x
 *         cursor_rect.y = y             # <<<<<<<<<<<<<<
 * 
 *         characters = font.get_characters()
 */
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_y); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 993, __pyx_L1_error)
  __pyx_v_cursor_rect.y = __pyx_t_7;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":995
 *         cursor_rect.y = y
 * 
 *         characters = font.get_characters()             # <<<<<<<<<<<<<<
 *         kernings = font.get_kernings()
 *         use_kerning = container.options['font_kerning']
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_font, __pyx_n_s_get_characters); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_characters, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":996
 * 
 *         characters = font.get_characters()
 *         kernings = font.get_kernings()             # <<<<<<<<<<<<<<
 *         use_kerning = container.options['font_kerning']
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_font, __pyx_n_s_get_kernings); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 996, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_kernings, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":997
 *         characters = font.get_characters()
 *         kernings = font.get_kernings()
 *         use_kerning = container.options['font_kerning']             # <<<<<<<<<<<<<<
 * 
 *         # Copy each character in the text string from the bitmap font atlas to the output texture/surface
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_container, __pyx_n_s_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_font_kerning); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_use_kerning = __pyx_t_4;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1000
 * 
 *         # Copy each character in the text string from the bitmap font atlas to the output texture/surface
 *         for index, text_char in enumerate(text):             # <<<<<<<<<<<<<<
 *             current_char = ord(text_char)
 *             if current_char in characters:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_2 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_text)) || PyTuple_CheckExact(__pyx_v_text)) {
    __pyx_t_1 = __pyx_v_text; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_text); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1000, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(1, 1000, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1000, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(1, 1000, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1000, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_9(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 1000, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_text_char, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1001
 *         # Copy each character in the text string from the bitmap font atlas to the output texture/surface
 *         for index, text_char in enumerate(text):
 *             current_char = ord(text_char)             # <<<<<<<<<<<<<<
 *             if current_char in characters:
 *                 char_info = characters[current_char]
 */
    __pyx_t_10 = __Pyx_PyObject_Ord(__pyx_v_text_char); if (unlikely(__pyx_t_10 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(1, 1001, __pyx_L1_error)
    __pyx_v_current_char = __pyx_t_10;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1002
 *         for index, text_char in enumerate(text):
 *             current_char = ord(text_char)
 *             if current_char in characters:             # <<<<<<<<<<<<<<
 *                 char_info = characters[current_char]
 * 
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_current_char); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_characters == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(1, 1002, __pyx_L1_error)
    }
    __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_t_3, __pyx_v_characters, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1002, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1003
 *             current_char = ord(text_char)
 *             if current_char in characters:
 *                 char_info = characters[current_char]             # <<<<<<<<<<<<<<
 * 
 *                 # need to ensure the first character is drawn at position 0 (adjust for initial xoffset value)
 */
      if (unlikely(__pyx_v_characters == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(1, 1003, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_current_char); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1003, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_characters, __pyx_t_3); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1003, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_char_info, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1006
 * 
 *                 # need to ensure the first character is drawn at position 0 (adjust for initial xoffset value)
 *                 if index == 0:             # <<<<<<<<<<<<<<
 *                     cursor_rect.x -= char_info.xoffset
 * 
 */
      __pyx_t_11 = __Pyx_PyInt_EqObjC(__pyx_v_index, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1006, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(1, 1006, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (__pyx_t_5) {

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1007
 *                 # need to ensure the first character is drawn at position 0 (adjust for initial xoffset value)
 *                 if index == 0:
 *                     cursor_rect.x -= char_info.xoffset             # <<<<<<<<<<<<<<
 * 
 *                 # apply kerning (if applicable)
 */
        __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_cursor_rect.x); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1007, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_char_info, __pyx_n_s_xoffset); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1007, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_12 = PyNumber_InPlaceSubtract(__pyx_t_11, __pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1007, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_12); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 1007, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_v_cursor_rect.x = __pyx_t_7;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1006
 * 
 *                 # need to ensure the first character is drawn at position 0 (adjust for initial xoffset value)
 *                 if index == 0:             # <<<<<<<<<<<<<<
 *                     cursor_rect.x -= char_info.xoffset
 * 
 */
      }

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1010
 * 
 *                 # apply kerning (if applicable)
 *                 if use_kerning and previous_char in kernings and current_char in kernings[previous_char]:             # <<<<<<<<<<<<<<
 *                     cursor_rect.x += kernings[previous_char][current_char]
 * 
 */
      __pyx_t_4 = (__pyx_v_use_kerning != 0);
      if (__pyx_t_4) {
      } else {
        __pyx_t_5 = __pyx_t_4;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_previous_char); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      if (unlikely(__pyx_v_kernings == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(1, 1010, __pyx_L1_error)
      }
      __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_t_12, __pyx_v_kernings, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1010, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_13 = (__pyx_t_4 != 0);
      if (__pyx_t_13) {
      } else {
        __pyx_t_5 = __pyx_t_13;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_current_char); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      if (unlikely(__pyx_v_kernings == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(1, 1010, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_previous_char); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_kernings, __pyx_t_3); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_13 = (__Pyx_PySequence_ContainsTF(__pyx_t_12, __pyx_t_11, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(1, 1010, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_4 = (__pyx_t_13 != 0);
      __pyx_t_5 = __pyx_t_4;
      __pyx_L11_bool_binop_done:;
      if (__pyx_t_5) {

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1011
 *                 # apply kerning (if applicable)
 *                 if use_kerning and previous_char in kernings and current_char in kernings[previous_char]:
 *                     cursor_rect.x += kernings[previous_char][current_char]             # <<<<<<<<<<<<<<
 * 
 *                 # set rectangles for character image copy and perform image copy operation
 */
        __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_cursor_rect.x); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1011, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (unlikely(__pyx_v_kernings == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(1, 1011, __pyx_L1_error)
        }
        __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_previous_char); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1011, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_kernings, __pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1011, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_current_char, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1011, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1011, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 1011, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_cursor_rect.x = __pyx_t_7;

        /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1010
 * 
 *                 # apply kerning (if applicable)
 *                 if use_kerning and previous_char in kernings and current_char in kernings[previous_char]:             # <<<<<<<<<<<<<<
 *                     cursor_rect.x += kernings[previous_char][current_char]
 * 
 */
      }

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1014
 * 
 *                 # set rectangles for character image copy and perform image copy operation
 *                 source_rect = char_info.rect             # <<<<<<<<<<<<<<
 *                 dest_rect.x = cursor_rect.x + char_info.xoffset
 *                 dest_rect.y = cursor_rect.y + char_info.yoffset
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_char_info, __pyx_n_s_rect); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = __pyx_convert__from_py_struct__SDL_Rect(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_source_rect = __pyx_t_14;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1015
 *                 # set rectangles for character image copy and perform image copy operation
 *                 source_rect = char_info.rect
 *                 dest_rect.x = cursor_rect.x + char_info.xoffset             # <<<<<<<<<<<<<<
 *                 dest_rect.y = cursor_rect.y + char_info.yoffset
 *                 SDL_BlitSurface(source_image, &source_rect, self.surface, &dest_rect)
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cursor_rect.x); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1015, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_char_info, __pyx_n_s_xoffset); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1015, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_11 = PyNumber_Add(__pyx_t_3, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1015, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_11); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 1015, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_v_dest_rect.x = __pyx_t_7;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1016
 *                 source_rect = char_info.rect
 *                 dest_rect.x = cursor_rect.x + char_info.xoffset
 *                 dest_rect.y = cursor_rect.y + char_info.yoffset             # <<<<<<<<<<<<<<
 *                 SDL_BlitSurface(source_image, &source_rect, self.surface, &dest_rect)
 * 
 */
      __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_cursor_rect.y); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1016, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_char_info, __pyx_n_s_yoffset); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1016, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_3 = PyNumber_Add(__pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1016, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 1016, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_dest_rect.y = __pyx_t_7;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1017
 *                 dest_rect.x = cursor_rect.x + char_info.xoffset
 *                 dest_rect.y = cursor_rect.y + char_info.yoffset
 *                 SDL_BlitSurface(source_image, &source_rect, self.surface, &dest_rect)             # <<<<<<<<<<<<<<
 * 
 *                 # advance the cursor in preparation for drawing the next character
 */
      (void)(SDL_BlitSurface(__pyx_v_source_image, (&__pyx_v_source_rect), __pyx_v_self->surface, (&__pyx_v_dest_rect)));

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1020
 * 
 *                 # advance the cursor in preparation for drawing the next character
 *                 cursor_rect.x += char_info.xadvance             # <<<<<<<<<<<<<<
 * 
 *             previous_char = current_char
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cursor_rect.x); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_char_info, __pyx_n_s_xadvance); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_11 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_11); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 1020, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_v_cursor_rect.x = __pyx_t_7;

      /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1002
 *         for index, text_char in enumerate(text):
 *             current_char = ord(text_char)
 *             if current_char in characters:             # <<<<<<<<<<<<<<
 *                 char_info = characters[current_char]
 * 
 */
    }

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1022
 *                 cursor_rect.x += char_info.xadvance
 * 
 *             previous_char = current_char             # <<<<<<<<<<<<<<
 * 
 *     def get_data(self):
 */
    __pyx_v_previous_char = __pyx_v_current_char;

    /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1000
 * 
 *         # Copy each character in the text string from the bitmap font atlas to the output texture/surface
 *         for index, text_char in enumerate(text):             # <<<<<<<<<<<<<<
 *             current_char = ord(text_char)
 *             if current_char in characters:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":947
 *             self.surface = NULL
 * 
 *     def render(self, container, text, x, y):             # <<<<<<<<<<<<<<
 *         """Render the text at the specified location."""
 *         cdef SDL_Rect cursor_rect
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer.render", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_characters);
  __Pyx_XDECREF(__pyx_v_kernings);
  __Pyx_XDECREF(__pyx_v_asset);
  __Pyx_XDECREF(__pyx_v_font);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XDECREF(__pyx_v_text_char);
  __Pyx_XDECREF(__pyx_v_char_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1024
 *             previous_char = current_char
 * 
 *     def get_data(self):             # <<<<<<<<<<<<<<
 *         """Return the bitmap font surface as ImageData (pixels)."""
 *         cdef int datalen = self.surface.w * self.surface.h * 4
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_9get_data(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_8get_data[] = "_SurfaceContainer.get_data(self)\nReturn the bitmap font surface as ImageData (pixels).";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_9get_data(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_data (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_8get_data(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_8get_data(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self) {
  int __pyx_v_datalen;
  PyObject *__pyx_v_pixels = 0;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_data", 0);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1026
 *     def get_data(self):
 *         """Return the bitmap font surface as ImageData (pixels)."""
 *         cdef int datalen = self.surface.w * self.surface.h * 4             # <<<<<<<<<<<<<<
 *         cdef bytes pixels = (<char *>self.surface.pixels)[:datalen]
 *         data = ImageData(self.w, self.h, 'rgba', pixels)
 */
  __pyx_v_datalen = ((__pyx_v_self->surface->w * __pyx_v_self->surface->h) * 4);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1027
 *         """Return the bitmap font surface as ImageData (pixels)."""
 *         cdef int datalen = self.surface.w * self.surface.h * 4
 *         cdef bytes pixels = (<char *>self.surface.pixels)[:datalen]             # <<<<<<<<<<<<<<
 *         data = ImageData(self.w, self.h, 'rgba', pixels)
 *         return data
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((char *)__pyx_v_self->surface->pixels) + 0, __pyx_v_datalen - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pixels = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1028
 *         cdef int datalen = self.surface.w * self.surface.h * 4
 *         cdef bytes pixels = (<char *>self.surface.pixels)[:datalen]
 *         data = ImageData(self.w, self.h, 'rgba', pixels)             # <<<<<<<<<<<<<<
 *         return data
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ImageData); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->w); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->h); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_t_3, __pyx_t_4, __pyx_n_u_rgba, __pyx_v_pixels};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1028, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_t_3, __pyx_t_4, __pyx_n_u_rgba, __pyx_v_pixels};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1028, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1028, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_4);
    __Pyx_INCREF(__pyx_n_u_rgba);
    __Pyx_GIVEREF(__pyx_n_u_rgba);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_n_u_rgba);
    __Pyx_INCREF(__pyx_v_pixels);
    __Pyx_GIVEREF(__pyx_v_pixels);
    PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_6, __pyx_v_pixels);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1028, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1029
 *         cdef bytes pixels = (<char *>self.surface.pixels)[:datalen]
 *         data = ImageData(self.w, self.h, 'rgba', pixels)
 *         return data             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_data);
  __pyx_r = __pyx_v_data;
  goto __pyx_L0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1024
 *             previous_char = current_char
 * 
 *     def get_data(self):             # <<<<<<<<<<<<<<
 *         """Return the bitmap font surface as ImageData (pixels)."""
 *         cdef int datalen = self.surface.w * self.surface.h * 4
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer.get_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pixels);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_10__reduce_cython__[] = "_SurfaceContainer.__reduce_cython__(self)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_10__reduce_cython__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_12__setstate_cython__[] = "_SurfaceContainer.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_12__setstate_cython__(((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct__SDL_Rect")
 * cdef struct_type __pyx_convert__from_py_struct__SDL_Rect(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

static struct SDL_Rect __pyx_convert__from_py_struct__SDL_Rect(PyObject *__pyx_v_obj) {
  struct SDL_Rect __pyx_v_result;
  PyObject *__pyx_v_value = NULL;
  struct SDL_Rect __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert__from_py_struct__SDL_Rect", 0);

  /* "FromPyStructUtility":13
 * cdef struct_type __pyx_convert__from_py_struct__SDL_Rect(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 */
  __pyx_t_1 = ((!(PyMapping_Check(__pyx_v_obj) != 0)) != 0);
  if (__pyx_t_1) {

    /* "FromPyStructUtility":14
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_2 = PyErr_Format(__pyx_builtin_TypeError, ((char const *)"Expected %.16s, got %.200s"), ((char *)"a mapping"), Py_TYPE(__pyx_v_obj)->tp_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "FromPyStructUtility":13
 * cdef struct_type __pyx_convert__from_py_struct__SDL_Rect(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 */
  }

  /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['x']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":17
 * 
 *     try:
 *         value = obj['x']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'x'")
 */
      __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_value = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['x']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "FromPyStructUtility":18
 *     try:
 *         value = obj['x']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'x'")
 *     result.x = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct__SDL_Rect", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 18, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "FromPyStructUtility":19
 *         value = obj['x']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'x'")             # <<<<<<<<<<<<<<
 *     result.x = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 19, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 19, __pyx_L6_except_error)
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['x']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L9_try_end:;
  }

  /* "FromPyStructUtility":20
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'x'")
 *     result.x = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['y']
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 20, __pyx_L1_error)
  __pyx_v_result.x = __pyx_t_6;

  /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'x'")
 *     result.x = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['y']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":22
 *     result.x = value
 *     try:
 *         value = obj['y']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'y'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_y); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 22, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'x'")
 *     result.x = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['y']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":23
 *     try:
 *         value = obj['y']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'y'")
 *     result.y = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct__SDL_Rect", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 23, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_2);

      /* "FromPyStructUtility":24
 *         value = obj['y']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'y'")             # <<<<<<<<<<<<<<
 *     result.y = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 24, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 24, __pyx_L14_except_error)
    }
    goto __pyx_L14_except_error;
    __pyx_L14_except_error:;

    /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'x'")
 *     result.x = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['y']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L17_try_end:;
  }

  /* "FromPyStructUtility":25
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'y'")
 *     result.y = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['w']
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 25, __pyx_L1_error)
  __pyx_v_result.y = __pyx_t_6;

  /* "FromPyStructUtility":26
 *         raise ValueError("No value specified for struct attribute 'y'")
 *     result.y = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['w']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":27
 *     result.y = value
 *     try:
 *         value = obj['w']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'w'")
 */
      __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_w); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L20_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "FromPyStructUtility":26
 *         raise ValueError("No value specified for struct attribute 'y'")
 *     result.y = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['w']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L25_try_end;
    __pyx_L20_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":28
 *     try:
 *         value = obj['w']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'w'")
 *     result.w = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct__SDL_Rect", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 28, __pyx_L22_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "FromPyStructUtility":29
 *         value = obj['w']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'w'")             # <<<<<<<<<<<<<<
 *     result.w = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 29, __pyx_L22_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 29, __pyx_L22_except_error)
    }
    goto __pyx_L22_except_error;
    __pyx_L22_except_error:;

    /* "FromPyStructUtility":26
 *         raise ValueError("No value specified for struct attribute 'y'")
 *     result.y = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['w']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L25_try_end:;
  }

  /* "FromPyStructUtility":30
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'w'")
 *     result.w = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['h']
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 30, __pyx_L1_error)
  __pyx_v_result.w = __pyx_t_6;

  /* "FromPyStructUtility":31
 *         raise ValueError("No value specified for struct attribute 'w'")
 *     result.w = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['h']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":32
 *     result.w = value
 *     try:
 *         value = obj['h']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'h'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_h); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 32, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":31
 *         raise ValueError("No value specified for struct attribute 'w'")
 *     result.w = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['h']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L33_try_end;
    __pyx_L28_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "FromPyStructUtility":33
 *     try:
 *         value = obj['h']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'h'")
 *     result.h = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct__SDL_Rect", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 33, __pyx_L30_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_2);

      /* "FromPyStructUtility":34
 *         value = obj['h']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'h'")             # <<<<<<<<<<<<<<
 *     result.h = value
 *     return result
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 34, __pyx_L30_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 34, __pyx_L30_except_error)
    }
    goto __pyx_L30_except_error;
    __pyx_L30_except_error:;

    /* "FromPyStructUtility":31
 *         raise ValueError("No value specified for struct attribute 'w'")
 *     result.w = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['h']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L33_try_end:;
  }

  /* "FromPyStructUtility":35
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'h'")
 *     result.h = value             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 35, __pyx_L1_error)
  __pyx_v_result.h = __pyx_t_6;

  /* "FromPyStructUtility":36
 *         raise ValueError("No value specified for struct attribute 'h'")
 *     result.h = value
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct__SDL_Rect")
 * cdef struct_type __pyx_convert__from_py_struct__SDL_Rect(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct__SDL_Rect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2id_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_rect(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_rect(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4rect_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_xoffset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_xoffset(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7xoffset_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_yoffset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_yoffset(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_7yoffset_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_xadvance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_xadvance(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_8xadvance_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_3__reduce_cython__, METH_NOARGS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_5__setstate_cython__, METH_O, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_4__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter[] = {
  {(char *)"id", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_id, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_id, (char *)"id: 'int'", 0},
  {(char *)"rect", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_rect, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_rect, (char *)"rect: 'SDL_Rect'", 0},
  {(char *)"xoffset", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_xoffset, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_xoffset, (char *)"xoffset: 'int'", 0},
  {(char *)"yoffset", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_yoffset, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_yoffset, (char *)"yoffset: 'int'", 0},
  {(char *)"xadvance", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_xadvance, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_19BitmapFontCharacter_xadvance, (char *)"xadvance: 'int'", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter = {
  PyVarObject_HEAD_INIT(0, 0)
  "mpfmc.uix.bitmap_font.bitmap_font.BitmapFontCharacter", /*tp_name*/
  sizeof(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)o);
  p->face = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->characters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->kernings = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont(PyObject *o) {
  struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *p = (struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_5__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->face);
  Py_CLEAR(p->characters);
  Py_CLEAR(p->kernings);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *p = (struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)o;
  if (p->characters) {
    e = (*v)(p->characters, a); if (e) return e;
  }
  if (p->kernings) {
    e = (*v)(p->kernings, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *p = (struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont *)o;
  tmp = ((PyObject*)p->characters);
  p->characters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->kernings);
  p->kernings = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_face(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_face(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_3__set__(o, v);
  }
  else {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4face_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_bold(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_bold(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4bold_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_italic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_italic(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6italic_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_padding(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_padding(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7padding_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_spacing(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_spacing(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7spacing_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_outline(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_outline(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7outline_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_line_height(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_line_height(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11line_height_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_base(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_4base_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_scale_w(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_scale_w(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_w_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_scale_h(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h_1__get__(o);
}

static int __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_scale_h(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7scale_h_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont[] = {
  {"get_image", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_7get_image, METH_NOARGS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_6get_image},
  {"get_characters", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_9get_characters, METH_NOARGS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_8get_characters},
  {"get_kernings", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_11get_kernings, METH_NOARGS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_10get_kernings},
  {"_load_descriptor_list", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_13_load_descriptor_list, METH_O, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_12_load_descriptor_list},
  {"_load_descriptor_file", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_15_load_descriptor_file, METH_O, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_14_load_descriptor_file},
  {"_load_descriptor_xml", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_17_load_descriptor_xml, METH_O, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_16_load_descriptor_xml},
  {"get_descent", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_19get_descent, METH_NOARGS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_18get_descent},
  {"get_ascent", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_21get_ascent, METH_NOARGS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_20get_ascent},
  {"get_extents", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_23get_extents, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_22get_extents},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_25__reduce_cython__, METH_NOARGS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_24__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_27__setstate_cython__, METH_O, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_26__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont[] = {
  {(char *)"face", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_face, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_face, (char *)"face: unicode", 0},
  {(char *)"bold", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_bold, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_bold, (char *)"bold: 'bool'", 0},
  {(char *)"italic", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_italic, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_italic, (char *)"italic: 'bool'", 0},
  {(char *)"padding", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_padding, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_padding, (char *)"padding: 'int'", 0},
  {(char *)"spacing", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_spacing, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_spacing, (char *)"spacing: 'int'", 0},
  {(char *)"outline", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_outline, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_outline, (char *)"outline: 'int'", 0},
  {(char *)"line_height", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_line_height, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_line_height, (char *)"line_height: 'int'", 0},
  {(char *)"base", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_base, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_base, (char *)"base: 'int'", 0},
  {(char *)"scale_w", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_scale_w, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_scale_w, (char *)"scale_w: 'int'", 0},
  {(char *)"scale_h", __pyx_getprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_scale_h, __pyx_setprop_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_scale_h, (char *)"scale_h: 'int'", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont = {
  PyVarObject_HEAD_INIT(0, 0)
  "mpfmc.uix.bitmap_font.bitmap_font.BitmapFont", /*tp_name*/
  sizeof(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "BitmapFont(unicode image_file, descriptor)", /*tp_doc*/
  __pyx_tp_traverse_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont, /*tp_traverse*/
  __pyx_tp_clear_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_10BitmapFont_3__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_5__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer[] = {
  {"render", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_7render, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_6render},
  {"get_data", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_9get_data, METH_NOARGS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_8get_data},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_11__reduce_cython__, METH_NOARGS, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_10__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_13__setstate_cython__, METH_O, __pyx_doc_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_12__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer = {
  PyVarObject_HEAD_INIT(0, 0)
  "mpfmc.uix.bitmap_font.bitmap_font._SurfaceContainer", /*tp_name*/
  sizeof(struct __pyx_obj_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "_SurfaceContainer(w, h)", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5mpfmc_3uix_11bitmap_font_11bitmap_font_17_SurfaceContainer_3__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_bitmap_font(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_bitmap_font},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "bitmap_font",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_BitmapFont, __pyx_k_BitmapFont, sizeof(__pyx_k_BitmapFont), 0, 0, 1, 1},
  {&__pyx_n_u_BitmapFont, __pyx_k_BitmapFont, sizeof(__pyx_k_BitmapFont), 0, 1, 0, 1},
  {&__pyx_n_s_BitmapFontCharacter, __pyx_k_BitmapFontCharacter, sizeof(__pyx_k_BitmapFontCharacter), 0, 0, 1, 1},
  {&__pyx_n_s_BitmapFontException, __pyx_k_BitmapFontException, sizeof(__pyx_k_BitmapFontException), 0, 0, 1, 1},
  {&__pyx_n_u_BitmapFontException, __pyx_k_BitmapFontException, sizeof(__pyx_k_BitmapFontException), 0, 1, 0, 1},
  {&__pyx_kp_u_Bitmap_font_descriptor_file_inva, __pyx_k_Bitmap_font_descriptor_file_inva, sizeof(__pyx_k_Bitmap_font_descriptor_file_inva), 0, 1, 0, 0},
  {&__pyx_kp_u_Bitmap_font_descriptor_file_inva_2, __pyx_k_Bitmap_font_descriptor_file_inva_2, sizeof(__pyx_k_Bitmap_font_descriptor_file_inva_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Could_not_load_bitmap_font_image, __pyx_k_Could_not_load_bitmap_font_image, sizeof(__pyx_k_Could_not_load_bitmap_font_image), 0, 1, 0, 0},
  {&__pyx_kp_u_Could_not_locate_the_bitmap_font, __pyx_k_Could_not_locate_the_bitmap_font, sizeof(__pyx_k_Could_not_locate_the_bitmap_font), 0, 1, 0, 0},
  {&__pyx_n_s_ElementTree, __pyx_k_ElementTree, sizeof(__pyx_k_ElementTree), 0, 0, 1, 1},
  {&__pyx_kp_s_Exception_returned_by_the_bitmap, __pyx_k_Exception_returned_by_the_bitmap, sizeof(__pyx_k_Exception_returned_by_the_bitmap), 0, 0, 1, 0},
  {&__pyx_n_s_IGNORECASE, __pyx_k_IGNORECASE, sizeof(__pyx_k_IGNORECASE), 0, 0, 1, 1},
  {&__pyx_kp_u_Illegal_value_in_bitmap_font_des, __pyx_k_Illegal_value_in_bitmap_font_des, sizeof(__pyx_k_Illegal_value_in_bitmap_font_des), 0, 1, 0, 0},
  {&__pyx_n_s_ImageData, __pyx_k_ImageData, sizeof(__pyx_k_ImageData), 0, 0, 1, 1},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_kp_s_No_value_specified_for_struct_at, __pyx_k_No_value_specified_for_struct_at, sizeof(__pyx_k_No_value_specified_for_struct_at), 0, 0, 1, 0},
  {&__pyx_kp_s_No_value_specified_for_struct_at_2, __pyx_k_No_value_specified_for_struct_at_2, sizeof(__pyx_k_No_value_specified_for_struct_at_2), 0, 0, 1, 0},
  {&__pyx_kp_s_No_value_specified_for_struct_at_3, __pyx_k_No_value_specified_for_struct_at_3, sizeof(__pyx_k_No_value_specified_for_struct_at_3), 0, 0, 1, 0},
  {&__pyx_kp_s_No_value_specified_for_struct_at_4, __pyx_k_No_value_specified_for_struct_at_4, sizeof(__pyx_k_No_value_specified_for_struct_at_4), 0, 0, 1, 0},
  {&__pyx_n_s_ParseError, __pyx_k_ParseError, sizeof(__pyx_k_ParseError), 0, 0, 1, 1},
  {&__pyx_n_s_SurfaceContainer, __pyx_k_SurfaceContainer, sizeof(__pyx_k_SurfaceContainer), 0, 0, 1, 1},
  {&__pyx_n_u_SurfaceContainer, __pyx_k_SurfaceContainer, sizeof(__pyx_k_SurfaceContainer), 0, 1, 0, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_u_amount, __pyx_k_amount, sizeof(__pyx_k_amount), 0, 1, 0, 1},
  {&__pyx_n_s_attrib, __pyx_k_attrib, sizeof(__pyx_k_attrib), 0, 0, 1, 1},
  {&__pyx_n_u_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 1, 0, 1},
  {&__pyx_kp_u_base_0_9_1_5, __pyx_k_base_0_9_1_5, sizeof(__pyx_k_base_0_9_1_5), 0, 1, 0, 0},
  {&__pyx_n_s_bitmap_font, __pyx_k_bitmap_font, sizeof(__pyx_k_bitmap_font), 0, 0, 1, 1},
  {&__pyx_n_u_char, __pyx_k_char, sizeof(__pyx_k_char), 0, 1, 0, 1},
  {&__pyx_kp_u_char_s_id_P_id_0_9_1_4_s_x_P_x_0, __pyx_k_char_s_id_P_id_0_9_1_4_s_x_P_x_0, sizeof(__pyx_k_char_s_id_P_id_0_9_1_4_s_x_P_x_0), 0, 1, 0, 0},
  {&__pyx_n_u_chars, __pyx_k_chars, sizeof(__pyx_k_chars), 0, 1, 0, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_u_common, __pyx_k_common, sizeof(__pyx_k_common), 0, 1, 0, 1},
  {&__pyx_n_s_container, __pyx_k_container, sizeof(__pyx_k_container), 0, 0, 1, 1},
  {&__pyx_n_s_descriptor, __pyx_k_descriptor, sizeof(__pyx_k_descriptor), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_find, __pyx_k_find, sizeof(__pyx_k_find), 0, 0, 1, 1},
  {&__pyx_n_s_findall, __pyx_k_findall, sizeof(__pyx_k_findall), 0, 0, 1, 1},
  {&__pyx_n_u_first, __pyx_k_first, sizeof(__pyx_k_first), 0, 1, 0, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_font_kerning, __pyx_k_font_kerning, sizeof(__pyx_k_font_kerning), 0, 0, 1, 1},
  {&__pyx_n_u_font_kerning, __pyx_k_font_kerning, sizeof(__pyx_k_font_kerning), 0, 1, 0, 1},
  {&__pyx_n_s_get_characters, __pyx_k_get_characters, sizeof(__pyx_k_get_characters), 0, 0, 1, 1},
  {&__pyx_n_s_get_font_asset, __pyx_k_get_font_asset, sizeof(__pyx_k_get_font_asset), 0, 0, 1, 1},
  {&__pyx_n_s_get_image, __pyx_k_get_image, sizeof(__pyx_k_get_image), 0, 0, 1, 1},
  {&__pyx_n_s_get_kernings, __pyx_k_get_kernings, sizeof(__pyx_k_get_kernings), 0, 0, 1, 1},
  {&__pyx_n_s_getroot, __pyx_k_getroot, sizeof(__pyx_k_getroot), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_group, __pyx_k_group, sizeof(__pyx_k_group), 0, 0, 1, 1},
  {&__pyx_n_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 1},
  {&__pyx_n_u_height, __pyx_k_height, sizeof(__pyx_k_height), 0, 1, 0, 1},
  {&__pyx_n_u_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 1, 0, 1},
  {&__pyx_n_s_image_file, __pyx_k_image_file, sizeof(__pyx_k_image_file), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_u_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 1, 0, 1},
  {&__pyx_n_s_isfile, __pyx_k_isfile, sizeof(__pyx_k_isfile), 0, 0, 1, 1},
  {&__pyx_n_u_kerning, __pyx_k_kerning, sizeof(__pyx_k_kerning), 0, 1, 0, 1},
  {&__pyx_kp_u_kerning_s_first_P_first_0_9_1_3, __pyx_k_kerning_s_first_P_first_0_9_1_3, sizeof(__pyx_k_kerning_s_first_P_first_0_9_1_3), 0, 1, 0, 0},
  {&__pyx_n_u_kernings, __pyx_k_kernings, sizeof(__pyx_k_kernings), 0, 1, 0, 1},
  {&__pyx_n_s_kivy_core_image, __pyx_k_kivy_core_image, sizeof(__pyx_k_kivy_core_image), 0, 0, 1, 1},
  {&__pyx_n_u_lineHeight, __pyx_k_lineHeight, sizeof(__pyx_k_lineHeight), 0, 1, 0, 1},
  {&__pyx_kp_u_lineHeight_0_9_1_5, __pyx_k_lineHeight_0_9_1_5, sizeof(__pyx_k_lineHeight_0_9_1_5), 0, 1, 0, 0},
  {&__pyx_n_s_load_descriptor_file, __pyx_k_load_descriptor_file, sizeof(__pyx_k_load_descriptor_file), 0, 0, 1, 1},
  {&__pyx_n_s_load_descriptor_list, __pyx_k_load_descriptor_list, sizeof(__pyx_k_load_descriptor_list), 0, 0, 1, 1},
  {&__pyx_n_s_load_descriptor_xml, __pyx_k_load_descriptor_xml, sizeof(__pyx_k_load_descriptor_xml), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_mpfmc_uix_bitmap_font_bitmap_fon, __pyx_k_mpfmc_uix_bitmap_font_bitmap_fon, sizeof(__pyx_k_mpfmc_uix_bitmap_font_bitmap_fon), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_options, __pyx_k_options, sizeof(__pyx_k_options), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {&__pyx_n_s_rect, __pyx_k_rect, sizeof(__pyx_k_rect), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_u_rgba, __pyx_k_rgba, sizeof(__pyx_k_rgba), 0, 1, 0, 1},
  {&__pyx_n_s_search, __pyx_k_search, sizeof(__pyx_k_search), 0, 0, 1, 1},
  {&__pyx_n_u_second, __pyx_k_second, sizeof(__pyx_k_second), 0, 1, 0, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_startswith, __pyx_k_startswith, sizeof(__pyx_k_startswith), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_text, __pyx_k_text, sizeof(__pyx_k_text), 0, 0, 1, 1},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {&__pyx_n_u_width, __pyx_k_width, sizeof(__pyx_k_width), 0, 1, 0, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_u_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 1, 0, 1},
  {&__pyx_n_s_xadvance, __pyx_k_xadvance, sizeof(__pyx_k_xadvance), 0, 0, 1, 1},
  {&__pyx_n_u_xadvance, __pyx_k_xadvance, sizeof(__pyx_k_xadvance), 0, 1, 0, 1},
  {&__pyx_n_s_xml_etree_ElementTree, __pyx_k_xml_etree_ElementTree, sizeof(__pyx_k_xml_etree_ElementTree), 0, 0, 1, 1},
  {&__pyx_n_s_xoffset, __pyx_k_xoffset, sizeof(__pyx_k_xoffset), 0, 0, 1, 1},
  {&__pyx_n_u_xoffset, __pyx_k_xoffset, sizeof(__pyx_k_xoffset), 0, 1, 0, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_u_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 1, 0, 1},
  {&__pyx_n_s_yoffset, __pyx_k_yoffset, sizeof(__pyx_k_yoffset), 0, 0, 1, 1},
  {&__pyx_n_u_yoffset, __pyx_k_yoffset, sizeof(__pyx_k_yoffset), 0, 1, 0, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(1, 764, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(1, 904, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 18, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 19, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":764
 * 
 *         # Open the descriptor file
 *         with open(descriptor_file) as text_file:             # <<<<<<<<<<<<<<
 * 
 *             # Loop over all the rows in the descriptor file
 */
  __pyx_tuple__4 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "FromPyStructUtility":19
 *         value = obj['x']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'x'")             # <<<<<<<<<<<<<<
 *     result.x = value
 *     try:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "FromPyStructUtility":24
 *         value = obj['y']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'y'")             # <<<<<<<<<<<<<<
 *     result.y = value
 *     try:
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_2); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "FromPyStructUtility":29
 *         value = obj['w']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'w'")             # <<<<<<<<<<<<<<
 *     result.w = value
 *     try:
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_3); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "FromPyStructUtility":34
 *         value = obj['h']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'h'")             # <<<<<<<<<<<<<<
 *     result.h = value
 *     return result
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_4); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":4
 * #cython: embedsignature=True, language_level=3
 * 
 * __all__ = ('BitmapFont',             # <<<<<<<<<<<<<<
 *            'BitmapFontException',
 *            '_SurfaceContainer',
 */
  __pyx_tuple__13 = PyTuple_Pack(3, __pyx_n_u_BitmapFont, __pyx_n_u_BitmapFontException, __pyx_n_u_SurfaceContainer); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter) < 0) __PYX_ERR(1, 614, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter.tp_dictoffset && __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BitmapFontCharacter, (PyObject *)&__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter) < 0) __PYX_ERR(1, 614, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter) < 0) __PYX_ERR(1, 614, __pyx_L1_error)
  __pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter = &__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFontCharacter;
  if (PyType_Ready(&__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont) < 0) __PYX_ERR(1, 632, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont.tp_dictoffset && __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BitmapFont, (PyObject *)&__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont) < 0) __PYX_ERR(1, 632, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont) < 0) __PYX_ERR(1, 632, __pyx_L1_error)
  __pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont = &__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font_BitmapFont;
  if (PyType_Ready(&__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer) < 0) __PYX_ERR(1, 928, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer.tp_dictoffset && __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SurfaceContainer, (PyObject *)&__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer) < 0) __PYX_ERR(1, 928, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer) < 0) __PYX_ERR(1, 928, __pyx_L1_error)
  __pyx_ptype_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer = &__pyx_type_5mpfmc_3uix_11bitmap_font_11bitmap_font__SurfaceContainer;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initbitmap_font(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initbitmap_font(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_bitmap_font(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_bitmap_font(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_bitmap_font(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'bitmap_font' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_bitmap_font(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("bitmap_font", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_mpfmc__uix__bitmap_font__bitmap_font) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(1, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "mpfmc.uix.bitmap_font.bitmap_font")) {
      if (unlikely(PyDict_SetItemString(modules, "mpfmc.uix.bitmap_font.bitmap_font", __pyx_m) < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":4
 * #cython: embedsignature=True, language_level=3
 * 
 * __all__ = ('BitmapFont',             # <<<<<<<<<<<<<<
 *            'BitmapFontException',
 *            '_SurfaceContainer',
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_tuple__13) < 0) __PYX_ERR(1, 4, __pyx_L1_error)

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":11
 * cimport cpython.pycapsule as pycapsule
 * 
 * import re             # <<<<<<<<<<<<<<
 * from os import path
 * from xml.etree.ElementTree import ElementTree, ParseError
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_re, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_1) < 0) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":12
 * 
 * import re
 * from os import path             # <<<<<<<<<<<<<<
 * from xml.etree.ElementTree import ElementTree, ParseError
 * from libc.string cimport memset
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_path);
  __Pyx_GIVEREF(__pyx_n_s_path);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_path);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_os, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_path, __pyx_t_1) < 0) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":13
 * import re
 * from os import path
 * from xml.etree.ElementTree import ElementTree, ParseError             # <<<<<<<<<<<<<<
 * from libc.string cimport memset
 * from kivy.core.image import ImageData
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ElementTree);
  __Pyx_GIVEREF(__pyx_n_s_ElementTree);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ElementTree);
  __Pyx_INCREF(__pyx_n_s_ParseError);
  __Pyx_GIVEREF(__pyx_n_s_ParseError);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_ParseError);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_xml_etree_ElementTree, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ElementTree); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ElementTree, __pyx_t_2) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ParseError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParseError, __pyx_t_2) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":15
 * from xml.etree.ElementTree import ElementTree, ParseError
 * from libc.string cimport memset
 * from kivy.core.image import ImageData             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_ImageData);
  __Pyx_GIVEREF(__pyx_n_s_ImageData);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ImageData);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_kivy_core_image, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ImageData); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ImageData, __pyx_t_1) < 0) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":609
 * 
 * 
 * class BitmapFontException(Exception):             # <<<<<<<<<<<<<<
 *     """Exception returned by the bitmap font module"""
 *     pass
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_BitmapFontException, __pyx_n_s_BitmapFontException, (PyObject *) NULL, __pyx_n_s_mpfmc_uix_bitmap_font_bitmap_fon, __pyx_kp_s_Exception_returned_by_the_bitmap); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_BitmapFontException, __pyx_t_2, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BitmapFontException, __pyx_t_4) < 0) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "mpfmc/uix/bitmap_font/bitmap_font.pyx":1
 * #!python             # <<<<<<<<<<<<<<
 * #cython: embedsignature=True, language_level=3
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct__SDL_Rect")
 * cdef struct_type __pyx_convert__from_py_struct__SDL_Rect(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init mpfmc.uix.bitmap_font.bitmap_font", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init mpfmc.uix.bitmap_font.bitmap_font");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* UnicodeAsUCS4 */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject* x) {
   Py_ssize_t length;
   #if CYTHON_PEP393_ENABLED
   length = PyUnicode_GET_LENGTH(x);
   if (likely(length == 1)) {
       return PyUnicode_READ_CHAR(x, 0);
   }
   #else
   length = PyUnicode_GET_SIZE(x);
   if (likely(length == 1)) {
       return PyUnicode_AS_UNICODE(x)[0];
   }
   #if Py_UNICODE_SIZE == 2
   else if (PyUnicode_GET_SIZE(x) == 2) {
       Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
       if (high_val >= 0xD800 && high_val <= 0xDBFF) {
           Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
           if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
               return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
           }
       }
   }
   #endif
   #endif
   PyErr_Format(PyExc_ValueError,
                "only single character unicode strings can be converted to Py_UCS4, "
                "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
   return (Py_UCS4)-1;
}

/* object_ord */
static long __Pyx__PyObject_Ord(PyObject* c) {
    Py_ssize_t size;
    if (PyBytes_Check(c)) {
        size = PyBytes_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyBytes_AS_STRING(c)[0];
        }
#if PY_MAJOR_VERSION < 3
    } else if (PyUnicode_Check(c)) {
        return (long)__Pyx_PyUnicode_AsPy_UCS4(c);
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    } else if (PyByteArray_Check(c)) {
        size = PyByteArray_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyByteArray_AS_STRING(c)[0];
        }
#endif
    } else {
        PyErr_Format(PyExc_TypeError,
            "ord() expected string of length 1, but %.200s found", c->ob_type->tp_name);
        return (long)(Py_UCS4)-1;
    }
    PyErr_Format(PyExc_TypeError,
        "ord() expected a character, but string of length %zd found", size);
    return (long)(Py_UCS4)-1;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* unicode_iter */
static CYTHON_INLINE int __Pyx_init_unicode_iteration(
    PyObject* ustring, Py_ssize_t *length, void** data, int *kind) {
#if CYTHON_PEP393_ENABLED
    if (unlikely(__Pyx_PyUnicode_READY(ustring) < 0)) return -1;
    *kind   = PyUnicode_KIND(ustring);
    *length = PyUnicode_GET_LENGTH(ustring);
    *data   = PyUnicode_DATA(ustring);
#else
    *kind   = 0;
    *length = PyUnicode_GET_SIZE(ustring);
    *data   = (void*)PyUnicode_AS_UNICODE(ustring);
#endif
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

static PyObject* __pyx_convert__to_py_struct__SDL_Rect(struct SDL_Rect s) {
  PyObject* res;
  PyObject* member;
  res = __Pyx_PyDict_NewPresized(4); if (unlikely(!res)) return NULL;
  member = __Pyx_PyInt_From_int(s.x); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_x, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_int(s.y); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_y, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_int(s.w); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_w, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_int(s.h); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_h, member) < 0)) goto bad;
  Py_DECREF(member);
  return res;
  bad:
  Py_XDECREF(member);
  Py_DECREF(res);
  return NULL;
}
/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
